"""
pyvr calibrate.

Usage:
    pyvr calibrate [options]

Options:
   -h, --help
   -c, --camera <camera>        Source of the camera to use for calibration [default: 0]
   -r, --resolution <res>       Input resolution in width and height [default: -1x-1]
   -n, --n_masks <n_masks>      Number of masks to calibrate [default: 1]
   -l, --load_from_file <file>  Load previous calibration settings [default: ranges.pickle]
   -s, --save <file>            Save calibration settings to a file [default: ranges.pickle]
"""

import logging
import pickle
import sys
from copy import copy
from pathlib import Path
from typing import Optional, List

import cv2
from docopt import docopt

from virtualreality import __version__


class ColorRange(object):
    def __init__(self,
                 color_num,
                 hue_center=0,
                 hue_range=180,
                 sat_center=0,
                 sat_range=180,
                 val_center=0,
                 val_range=180
                 ):
        self.color_num = color_num
        self.hue_center = hue_center
        self.hue_range = hue_range
        self.sat_center = sat_center
        self.sat_range = sat_range
        self.val_center = val_center
        self.val_range = val_range


class CalibrationData(object):
    def __init__(self, width=1, height=1, auto_exposure=0.25, exposure=0, saturation=50, num_colors=4):
        self.width = width
        self.height = height
        self.exposure = exposure
        self.saturation = saturation
        self.num_colors = num_colors
        self.color_ranges: List[ColorRange] = []

        color_dist = 180 // num_colors
        for color in range(num_colors):
            self.color_ranges.append(ColorRange(color, *[color * color_dist, color_dist] * 3))

    @classmethod
    def load_from_file(cls, load_file: str = str(Path(__file__).parent) + "ranges.pickle") -> Optional[
        'CalibrationData']:
        """Load the calibration data from a file."""
        try:
            with open(load_file, "rb") as file:
                ranges = pickle.load(file)
            return ranges
        except FileNotFoundError as fe:
            logging.warning(f"Could not load calibration file '{load_file}'.")

    def save_to_file(self, save_file: str = str(Path(__file__).parent) + "ranges.pickle") -> None:
        with open(save_file, "wb") as file:
            pickle.dump(self, file)

def colordata_to_blob(colordata, mapdata):
    '''
    translates CalibrationData object to BlobTracker format masks

    :colordata: CalibrationData object
    :mapdata: a map dict with key representing the mask name and value representing the mask number

    '''
    out = {}

    for key, clr_range_index in mapdata.items():
        temp = colordata.color_ranges[clr_range_index]
        out[key] = {
                'h':(temp.hue_center, temp.hue_range),
                's':(temp.sat_center, temp.sat_range),
                'v':(temp.val_center, temp.val_range),
                    }

    return out

def load_mapdata_from_file(path):
    '''
    loads mapdata from file, for use in colordata_to_blob
    '''
    with open(path, 'rb') as file:
        return pickle.load(file)

def save_mapdata_to_file(path, mapdata):
    '''
    save mapdata to file, for use in colordata_to_blob
    '''
    with open(path, "wb") as file:
        pickle.dump(mapdata, file)

def list_supported_capture_properties(cap: cv2.VideoCapture):
    """List the properties supported by the capture device."""
    # thanks: https://stackoverflow.com/q/47935846/782170
    supported = list()
    for attr in dir(cv2):
        if attr.startswith("CAP_PROP") and cap.get(getattr(cv2, attr)) != -1:
            supported.append(attr)
    return supported


def get_color_mask(hsv, color_range: ColorRange):
    color_low = [
        color_range.hue_center - color_range.hue_range,
        color_range.sat_center - color_range.sat_range,
        color_range.val_center - color_range.val_range,
    ]

    color_high = [
        color_range.hue_center + color_range.hue_range,
        color_range.sat_center + color_range.sat_range,
        color_range.val_center + color_range.val_range,
    ]

    color_low_neg = copy(color_low)
    color_high_neg = copy(color_high)
    for c in range(3):
        if c==0:
            c_max = 180
        else:
            c_max = 255
        if color_low_neg[c] < 0:
            color_low_neg[c] = c_max + color_low_neg[c]
            color_high_neg[c] = c_max
            color_low[c] = 0
        elif color_high_neg[c] > c_max:
            color_low_neg[c] = 0
            color_high_neg[c] = color_high_neg[c] - c_max
            color_high[c] = c_max

    mask1 = cv2.inRange(hsv, tuple(color_low), tuple(color_high))
    mask2 = cv2.inRange(hsv, tuple(color_low_neg), tuple(color_high_neg))
    mask = cv2.bitwise_or(mask1, mask2)
    return mask


def _set_default_camera_properties(vs, cam, vs_supported, frame_width, frame_height):
    if "CAP_PROP_FOURCC" not in vs_supported:
        logging.warning(f"Camera {cam} does not support setting video codec.")
    else:
        vs.set(cv2.CAP_PROP_FOURCC, cv2.CAP_OPENCV_MJPEG)

    if "CAP_PROP_AUTO_EXPOSURE" not in vs_supported:
        logging.warning(f"Camera {cam} does not support turning on/off auto exposure.")
    else:
        vs.set(cv2.CAP_PROP_AUTO_EXPOSURE, 0.25)

    if "CAP_PROP_EXPOSURE" not in vs_supported:
        logging.warning(f"Camera {cam} does not support directly setting exposure.")
    else:
        vs.set(cv2.CAP_PROP_EXPOSURE, -7)

    if "CAP_PROP_EXPOSURE" not in vs_supported:
        logging.warning(f"Camera {cam} does not support directly setting exposure.")
    else:
        vs.set(cv2.CAP_PROP_EXPOSURE, -7)

    if "CAP_PROP_FRAME_HEIGHT" not in vs_supported:
        logging.warning(f"Camera {cam} does not support requesting frame height.")
    else:
        vs.set(cv2.CAP_PROP_FRAME_HEIGHT, frame_height)

    if "CAP_PROP_FRAME_WIDTH" not in vs_supported:
        logging.warning(f"Camera {cam} does not support requesting frame width.")
    else:
        vs.set(cv2.CAP_PROP_FRAME_WIDTH, frame_width)


def manual_calibration(
        cam=0, num_colors_to_track=4, frame_width=-1, frame_height=-1, load_file="", save_file="ranges.pickle"
):
    """Manually calibrate the hsv ranges and camera settings used for blob tracking."""
    vs = cv2.VideoCapture(cam)
    vs.set(cv2.CAP_PROP_EXPOSURE, -7)
    vs_supported = list_supported_capture_properties(vs)

    _set_default_camera_properties(vs, cam, vs_supported, frame_width, frame_height)

    cam_window = f"camera {cam} input"
    cv2.namedWindow(cam_window)
    if "CAP_PROP_EXPOSURE" in vs_supported:
        cv2.createTrackbar(
            "exposure", cam_window, 0, 16, lambda x: vs.set(cv2.CAP_PROP_EXPOSURE, x - 8),
        )
    if "CAP_PROP_SATURATION" in vs_supported:
        cv2.createTrackbar(
            "saturation", cam_window, 0, 100, lambda x: vs.set(cv2.CAP_PROP_SATURATION, x),
        )
    else:
        logging.warning(f"Camera {cam} does not support setting saturation.")

    ranges = None
    if load_file:
        ranges = CalibrationData.load_from_file(load_file)
    if ranges is None:
        ranges = CalibrationData(width=frame_width, height=frame_height, num_colors=num_colors_to_track)

    tracker_window_names = []
    for color in range(num_colors_to_track):
        tracker_window_names.append(f"color {color}")
        cv2.namedWindow(tracker_window_names[color])

        cv2.createTrackbar(
            "hue center", tracker_window_names[color], ranges.color_ranges[color].hue_center, 180, lambda _: None,
        )
        cv2.createTrackbar(
            "hue range", tracker_window_names[color], ranges.color_ranges[color].hue_range, 180, lambda _: None,
        )
        cv2.createTrackbar(
            "sat center", tracker_window_names[color], ranges.color_ranges[color].sat_center, 255, lambda _: None,
        )
        cv2.createTrackbar(
            "sat range", tracker_window_names[color], ranges.color_ranges[color].sat_range, 255, lambda _: None,
        )
        cv2.createTrackbar(
            "val center", tracker_window_names[color], ranges.color_ranges[color].val_center, 255, lambda _: None,
        )
        cv2.createTrackbar(
            "val range", tracker_window_names[color], ranges.color_ranges[color].val_range, 255, lambda _: None,
        )

    while 1:
        ret, frame = vs.read()

        if frame is None:
            break

        blurred = cv2.GaussianBlur(frame, (3, 3), 0)
        hsv = cv2.cvtColor(blurred, cv2.COLOR_BGR2HSV)

        exposure = cv2.getTrackbarPos("exposure", cam_window)
        saturation = cv2.getTrackbarPos("saturation", cam_window)

        ranges.exposure = exposure - 8
        ranges.saturation = saturation

        for color in range(num_colors_to_track):
            hue_center = cv2.getTrackbarPos("hue center", tracker_window_names[color])
            hue_range = cv2.getTrackbarPos("hue range", tracker_window_names[color])
            sat_center = cv2.getTrackbarPos("sat center", tracker_window_names[color])
            sat_range = cv2.getTrackbarPos("sat range", tracker_window_names[color])
            val_center = cv2.getTrackbarPos("val center", tracker_window_names[color])
            val_range = cv2.getTrackbarPos("val range", tracker_window_names[color])

            ranges.color_ranges[color].hue_center = hue_center
            ranges.color_ranges[color].hue_range = hue_range
            ranges.color_ranges[color].sat_center = sat_center
            ranges.color_ranges[color].sat_range = sat_range
            ranges.color_ranges[color].val_center = val_center
            ranges.color_ranges[color].val_range = val_range

            mask = get_color_mask(hsv, ranges.color_ranges[color])

            res = cv2.bitwise_and(hsv, hsv, mask=mask)

            cv2.imshow(tracker_window_names[color], res)

        cv2.imshow(cam_window, frame)

        k = cv2.waitKey(1) & 0xFF

        if k in [ord("q"), 27]:
            break

    for color in range(num_colors_to_track):
        hue_center = cv2.getTrackbarPos("hue center", tracker_window_names[color])
        hue_range = cv2.getTrackbarPos("hue range", tracker_window_names[color])
        sat_center = cv2.getTrackbarPos("sat center", tracker_window_names[color])
        sat_range = cv2.getTrackbarPos("sat range", tracker_window_names[color])
        val_center = cv2.getTrackbarPos("val center", tracker_window_names[color])
        val_range = cv2.getTrackbarPos("val range", tracker_window_names[color])

        print(f"hue_center[{color}]: {hue_center}")
        print(f"hue_range[{color}]: {hue_range}")
        print(f"sat_center[{color}]: {sat_center}")
        print(f"sat_range[{color}]: {sat_range}")
        print(f"val_center[{color}]: {val_center}")
        print(f"val_range[{color}]: {val_range}")

    if save_file:
        ranges.save_to_file(save_file)
        print(f'ranges saved to list in "{save_file}".')
        print("You can use this in the pyvr tracker using the --calibration-file argument.")

    vs.release()
    cv2.destroyAllWindows()


def main():
    """Calibrate entry point."""
    # allow calling from both python -m and from pyvr:
    argv = sys.argv[1:]
    if len(argv) < 2 or sys.argv[1] != "calibrate":
        argv = ["calibrate"] + argv

    args = docopt(__doc__, version=f"pyvr version {__version__}", argv=argv)

    width, height = args["--resolution"].split("x")

    if args["--camera"].isdigit():
        cam = int(args["--camera"])
    else:
        cam = args["--camera"]

    manual_calibration(
        cam=cam,
        num_colors_to_track=int(args["--n_masks"]),
        frame_width=int(width),
        frame_height=int(height),
        load_file=args["--load_from_file"],
        save_file=args["--save"],
    )



import bluetooth
import time

import threading


class BluetoothReceiver(threading.Thread):
    """
    no docs... too bad!

    oh yeah, here is an example:
        addr = '98:D3:51:FD:EB:C4' # some device address, whatever
        port = 1

        t = BluetoothReceiver(addr, port)

        with t:
            for _ in range(1000):
                print (t.last_read)
                time.sleep(1/100) # time delay, doesn't affect receiving, whatever whatever
    """

    def __init__(self, addr, port):
        """
        here is a riddle for you: what does init do?
        """
        super().__init__()
        self.backBuffer = bytearray()
        self._last_read = b""
        self._terminator = b"\n"

        self.sock = bluetooth.BluetoothSocket(bluetooth.RFCOMM)
        self.sock.connect((addr, port))

        self.readSize = 100

        # -------------------threading related------------------------

        self.alive = True
        self._lock = threading.Lock()
        self.daemon = False

    def __enter__(self):
        self.start()
        if not self.alive:
            self.close()
            raise RuntimeError("receive thread already finished")

    def __exit__(self, *exp):
        self.stop()

    def stop(self):
        self.alive = False
        self.join(4)
        time.sleep(0.01)
        self.close()

    def close(self):
        """hammer time!"""
        time.sleep(0.5)
        self.sock.close()

    @property
    def last_read(self):
        return self._last_read

    def send(self, data):
        """sends data or something idk"""
        self.sock.send(data)

    def run(self):
        while self.alive:
            try:
                data = self.sock.recv(self.readSize)
                self.backBuffer.extend(data)

                if not data:
                    break

                while self._terminator in self.backBuffer:
                    self._last_read, self.backBuffer = self.backBuffer.split(self._terminator, 1)

            except Exception as e:
                print(f"BluetoothSocket receive thread failed: {repr(e)}")
                break

        self.alive = False



import time

import numpy as np
from pykalman import KalmanFilter


class EulerKalman(object):
    """
    A Kalman filter that keeps track of euler position and rotation.

    example usage:
        >>> k = EulerKalman() # init filter

        for i in range(10):
            print (t.apply([5+i, 6+i, 7+i])) # apply and update filter
    """

    def __init__(self, init_state=tuple(0 for _ in range(18)), n_iter=5, train_size=15):
        """
        Create the Kalman filter.

        :param init_state: Initial state of the Kalman filter. Should be equal to first element in first_train_batch.
        :param n_iter: Number of times to repeat the parameter estimation function (estimates noise)
        """
        init_state = np.array(init_state)
        if init_state.size != 18:
            raise ValueError("EulerKalman expects init: [x,y,z,vx,vy,vz,ax,ay,az,r,p,w,vr,vp,vw,ar,ap,aw]")
        self.t_prev = time.time()

        transition_matrix = self._get_transition_matrix()
        observation_matrix = np.eye(18)

        self._expected_shape = init_state.shape

        self._filter = KalmanFilter(
            transition_matrices=transition_matrix,
            observation_matrices=observation_matrix,
            initial_state_mean=init_state,
        )

        self._calibration_countdown = 0
        self._calibration_observations = []
        self._em_iter = n_iter

        self.calibrate(train_size, n_iter)

        self._x_now = np.zeros(3)
        self._p_now = np.zeros(3)

    def _get_transition_matrix(self):
        """
        State transition matrix for euler position+rotation with velocity and acceleration.

        See: http://campar.in.tum.de/Chair/KalmanFilter#Linear_Kalman_Filter_for_the_pos

        Assuming dt is 0.1, this gives:
        1.0     0.0     0.0     0.1     0.0     0.0     0.005   0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0
        0.0     1.0     0.0     0.0     0.1     0.0     0.0     0.005   0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0
        0.0     0.0     1.0     0.0     0.0     0.1     0.0     0.0     0.005   0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0
        0.0     0.0     0.0     1.0     0.0     0.0     0.1     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0
        0.0     0.0     0.0     0.0     1.0     0.0     0.0     0.1     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0
        0.0     0.0     0.0     0.0     0.0     1.0     0.0     0.0     0.1     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0
        0.0     0.0     0.0     0.0     0.0     0.0     1.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0
        0.0     0.0     0.0     0.0     0.0     0.0     0.0     1.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0
        0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     1.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0
        0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     1.0     0.0     0.0     0.1     0.0     0.0     0.005   0.0     0.0
        0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     1.0     0.0     0.0     0.1     0.0     0.0     0.005   0.0
        0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     1.0     0.0     0.0     0.1     0.0     0.0     0.005
        0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     1.0     0.0     0.0     0.1     0.0     0.0
        0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     1.0     0.0     0.0     0.1     0.0
        0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     1.0     0.0     0.0     0.1
        0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     1.0     0.0     0.0
        0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     1.0     0.0
        0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     0.0     1.0
        :return:
        """
        a = np.eye(9)
        t_now = time.time()
        dt = t_now - self.t_prev
        a[tuple(i - 3 for i in range(3, 9)), tuple(i for i in range(3, 9))] = dt
        a[tuple(i - 6 for i in range(6, 9)), tuple(i for i in range(6, 9))] = 0.5 * dt ** 2
        m = np.zeros((18, 18))
        m[0:9, 0:9] = a
        m[9:18, 9:18] = a
        self.t_prev = t_now
        return m

    def _check_calibration(self, obz):
        if self._calibration_countdown:
            self._calibration_observations.append(obz)
            self._calibration_countdown -= 1
            if self._calibration_countdown == 0:
                self._run_calibration()
        return self._x_now

    def apply_imu(self, imu):
        self._filter.transition_matrices = self._get_transition_matrix()

        # run with no observation to get current data for info the imu doesn't have.
        temp_x_now, _ = self._filter.filter_update(
            filtered_state_mean=self._x_now, filtered_state_covariance=self._p_now
        )

        acc = imu.get_acc(q=temp_x_now[9:12])
        orient = imu.get_instant_orientation(accel=acc)
        gyro = imu.get_gyro()
        partial_obz = temp_x_now
        partial_obz[6:9] = acc
        partial_obz[9:12] = gyro
        partial_obz[15:18] = orient

        self._x_now, new_p_now = self._filter.filter_update(
            filtered_state_mean=self._x_now, filtered_state_covariance=self._p_now, observation=partial_obz,
        )

        # only update measured variances of things we actually measured
        self._p_now[6:9] = new_p_now[6:9]
        self._p_now[9:12] = new_p_now[9:12]
        self._p_now[15:18] = new_p_now[15:18]

        self._check_calibration(partial_obz)

        return self._x_now

    def apply_blob(self, xyz):
        """Apply the Kalman filter against the observation obz."""
        # Todo: automatically start a calibration routine if everything is at rest for long enough,
        #  or if the user leaves, or if there is enough error. Also, add some UI calibrate button, save, and load.
        assert len(xyz) == 3

        self._filter.transition_matrices = self._get_transition_matrix()

        # run with no observation to get current data for info the imu doesn't have.
        temp_x_now, _ = self._filter.filter_update(
            filtered_state_mean=self._x_now, filtered_state_covariance=self._p_now
        )

        partial_obz = temp_x_now
        partial_obz[0:3] = xyz

        self._x_now, new_p_now = self._filter.filter_update(
            filtered_state_mean=self._x_now, filtered_state_covariance=self._p_now, observation=partial_obz,
        )

        # only update measured variances of things we actually measured
        self._p_now[0:3] = new_p_now[0:3]

        self._check_calibration(partial_obz)

        return self._x_now

    def _run_calibration(self):
        """Update the Kalman filter so that noise matrices are more accurate."""
        t_start = time.time()
        self._filter = self._filter.em(self._calibration_observations, n_iter=self._em_iter)
        print(f" kalman filter calibrated, took {time.time() - t_start}s")
        f_means, f_covars = self._filter.filter(self._calibration_observations)
        self._x_now = f_means[-1, :]
        self._p_now = f_covars[-1, :]
        self._calibration_observations = []

    def calibrate(self, observations=100000, em_iter=10):
        """
        Start the calibration routine.

        :param observations: Number of observations before calibration runs.
        :param em_iter: number of times calibration runs over all observations.
        """
        self._calibration_countdown = observations
        self._em_iter = em_iter



import logging
import os
import enum

class log_levels(enum.Enum) :
    DEBUG=logging.DEBUG
    INFO=logging.INFO
    WARNING=logging.WARNING
    ERROR=logging.ERROR
    CRITICAL=logging.CRITICAL

def setup_custom_logger(name, level, file, format='[%(asctime)s] %(name)s %(levelname)s - %(message)s', console_logging=False) :
    logger = logging.getLogger(name)
    logger.setLevel(level)

    log_path = os.path.normpath(os.path.join(os.path.dirname(__file__), file))
    dir_p = os.path.dirname(log_path)
    if not os.path.exists(dir_p):
        os.mkdir(dir_p)

    #logging entry format
    logger_formatter = logging.Formatter('[%(asctime)s] %(name)s %(levelname)s - %(message)s')

    logger_handler = logging.FileHandler(log_path)
    logger_handler.setLevel(level)
    logger_handler.setFormatter(logger_formatter)
    logger.addHandler(logger_handler)

    if(console_logging):
        console_logger_handler = logging.StreamHandler()
        console_logger_handler.setLevel(level)
        console_logger_handler.setFormatter(logger_formatter)
        logger.addHandler(console_logger_handler)

    return logger



"""Various VR utilities."""

import queue
import threading
import time
from asyncio.streams import StreamReader
from typing import Sequence, Dict

import cv2
import numpy as np
import serial.threaded
from pykalman import KalmanFilter
from displayarray import read_updates

from itertools import islice, takewhile
import re


def format_str_for_write(input_str: str) -> bytes:
    """Format a string for writing to SteamVR's stream."""
    if len(input_str) < 1:
        return "".encode("utf-8")

    if input_str[-1] != "\n":
        return (input_str + "\n").encode("utf-8")

    return input_str.encode("utf-8")


async def read(reader: StreamReader, read_len: int = 20) -> str:
    """Read one line from reader asynchronously."""
    data = []
    temp = " "
    while "\n" not in temp and temp != "":
        temp = await reader.read(read_len)
        temp = temp.decode("utf-8")
        data.append(temp)
        time.sleep(0)  # allows thread switching

    return "".join(data)


def read2(reader, read_len=20):
    """Read one line from reader asynchronously."""
    data = []
    temp = " "
    while "\n" not in temp and temp != "":
        temp = reader.recv(read_len)
        temp = temp.decode("utf-8")
        data.append(temp)
        time.sleep(0)  # allows thread switching

    return "".join(data)

async def read3(reader: StreamReader, read_len: int = 20) -> str:
    """Read one line from reader asynchronously."""
    data = bytearray()
    temp = b" "
    while b"\n" not in temp and temp != b"":
        temp = await reader.read(read_len)
        data.extend(temp)
        # time.sleep(0)  # allows thread switching

    return data


def rotate_z(points: Dict[str, Dict[str, float]], angle: float):
    """Rotate a set of points around the z axis."""
    sin = np.sin(angle)
    cos = np.cos(angle)

    for key, p in points.items():
        points[key]["x"] = p["x"] * cos - p["y"] * sin
        points[key]["y"] = p["y"] * cos + p["x"] * sin


def rotate_x(points: Dict[str, Dict[str, float]], angle: float):
    """Rotate a set of points around the x axis."""
    sin = np.sin(angle)
    cos = np.cos(angle)

    for key, p in points.items():
        points[key]["y"] = p["y"] * cos - p["z"] * sin
        points[key]["z"] = p["z"] * cos + p["y"] * sin


def rotate_y(points: Dict[str, Dict[str, float]], angle: float):
    """Rotate a set of points around the y axis."""
    sin = np.sin(angle)
    cos = np.cos(angle)

    for key, p in points.items():
        points[key]["x"] = p["x"] * cos - p["z"] * sin
        points[key]["z"] = p["z"] * cos + p["x"] * sin


def rotate(points: Dict[str, Dict[str, float]], angles: Sequence[float]) -> None:
    """
    Rotate a set of points around the x, y, then z axes.

    :param points: a point dictionary, such as: `{"marker 1" : {"x": 0, "y": 0, "z": 0}}`
    :param angles: the degrees to rotate on the x, y, and z axes
    """
    if angles[0] != 0:
        rotate_x(points, angles[0])

    if angles[1] != 0:
        rotate_y(points, angles[1])

    if angles[2] != 0:
        rotate_z(points, angles[2])


def translate(points: Dict[str, Dict[str, float]], offsets: Sequence[float]) -> None:
    """
    Translate a set of points along the x, y, then z axes.

    :param points: a point dictionary, such as: `{"marker 1" : {"x": 0, "y": 0, "z": 0}}`
    :param offsets: the distance to translate on the x, y, and z axes
    """
    for key, p in points.items():
        points[key]["x"] += offsets[0]
        points[key]["y"] += offsets[1]
        points[key]["z"] += offsets[2]


def strings_share_characters(str1: str, str2: str) -> bool:
    """Determine if two strings share any characters."""
    for i in str2:
        if i in str1:
            return True

    return False


def get_numbers_from_text(text, separator="\t"):
    """Get a list of number from a string of numbers seperated by :separator:[default: "\t"]."""
    try:
        if isinstance(text, bytearray) or isinstance(text, bytes):
            text = text.decode("utf-8")

        if strings_share_characters(text.lower(), "qwrtyuiopsasdfghjklzxcvbnm><*[]{}()") or len(text) == 0:
            return []

        return [float(i) for i in text.split(separator)]

    except Exception as e:
        print(f"get_numbers_from_text: {repr(e)} {repr(text)}")

        return []

def get_pose_struct_from_text(text):
    '''returns struct from :text:, :text: has to be styled so that it matches this regex: ([htc][0-9]+[ ])+([htc][0-9]+)$|([htc][0-9]+)$'''
    res = re.search('([htc][0-9]+[ ])+([htc][0-9]+)$|([htc][0-9]+)$', text)
    if res != None:
        if res.group(0) == text:
            return tuple(i[0] for i in text.split(' ')), tuple(int(i[1:]) for i in text.split(' '))
    return (), ()

def parse_poses_from_packet(packet, struct):
    '''parses all poses from a :packet:, provided a packet :struct:'''
    it = iter(packet)
    return [tuple(islice(it, 0, i)) for i in struct]

def get_poses_shape(poses):
    '''returns a shape of :poses: parsed by parse_poses_from_packet'''
    return tuple(len(i) for i in poses)

def has_nan_in_pose(pose):
    """Determine if any numbers in pose are invalid."""
    for key in ["x", "y", "z"]:
        if np.isnan(pose[key]) or np.isinf(pose[key]):
            return True

    return False


class LazyKalman:
    """
    no docs... too bad.

    But then, SimLeek added some.

    example usage:
        t = LazyKalman([1, 2, 3], np.eye(3), np.eye(3)) # init filter

        for i in range(10):
            print (t.apply([5+i, 6+i, 7+i])) # apply and update filter
    """

    def __init__(self, init_state, transition_matrix, observation_matrix, n_iter=5, train_size=15):
        """
        Create the Kalman filter.

        :param init_state: Initial state of the Kalman filter. Should be equal to first element in first_train_batch.
        :param transition_matrix: adjacency matrix representing state transition from t to t+1 for any time t
                                  Example: http://campar.in.tum.de/Chair/KalmanFilter
                                  Most likely, this will be an NxN eye matrix the where N is the number of variables
                                  being estimated
        :param observation_matrix: translation matrix from measurement coordinate system to desired coordinate system
                                   See: https://dsp.stackexchange.com/a/27488
                                   Most likely, this will be an NxN eye matrix the where N is the number of variables
                                   being estimated
        :param n_iter: Number of times to repeat the parameter estimation function (estimates noise)
        """
        init_state = np.array(init_state)
        transition_matrix = np.array(transition_matrix)
        observation_matrix = np.array(observation_matrix)

        self._expected_shape = init_state.shape

        self._filter = KalmanFilter(
            transition_matrices=transition_matrix,
            observation_matrices=observation_matrix,
            initial_state_mean=init_state,
        )

        self._calibration_countdown = 0
        self._calibration_observations = []
        self._em_iter = n_iter

        self.calibrate(train_size, n_iter)

        self._x_now = np.zeros(3)
        self._p_now = np.zeros(3)

    def apply(self, obz):
        """Apply the Kalman filter against the observation obz."""
        # Todo: automatically start a calibration routine if everything is at rest for long enough,
        #  or if the user leaves, or if there is enough error. Also, add some UI calibrate button, save, and load.
        obz = np.array(obz)

        assert obz.shape == self._expected_shape, "shape miss match"

        self._x_now, self._p_now = self._filter.filter_update(
            filtered_state_mean=self._x_now, filtered_state_covariance=self._p_now, observation=obz,
        )

        if self._calibration_countdown:
            self._calibration_observations.append(obz)
            self._calibration_countdown -= 1
            if self._calibration_countdown == 0:
                self._run_calibration()
        return self._x_now

    def _run_calibration(self):
        """Update the Kalman filter so that noise matrices are more accurate."""
        t_start = time.time()
        self._filter = self._filter.em(self._calibration_observations, n_iter=self._em_iter)
        print(f" kalman filter calibrated, took {time.time() - t_start}s")
        f_means, f_covars = self._filter.filter(self._calibration_observations)
        self._x_now = f_means[-1, :]
        self._p_now = f_covars[-1, :]
        self._calibration_observations = []

    def calibrate(self, observations=100000, em_iter=10):
        """
        Start the calibration routine.

        :param observations: Number of observations before calibration runs.
        :param em_iter: number of times calibration runs over all observations.
        """
        self._calibration_countdown = observations
        self._em_iter = em_iter


class SerialReaderFactory(serial.threaded.LineReader):
    """
    A protocol factory for serial.threaded.ReaderThread.

    self.lastRead should be read only, if you need to modify it make a copy

    usage:
        with serial.threaded.ReaderThread(serial_instance, SerialReaderFactory) as protocol:
            protocol.lastRead # contains the last incoming message from serial
            protocol.write_line(single_line_text) # used to write a single line to serial
    """

    def __init__(self):
        """Create the SerialReaderFactory."""
        super().__init__()
        self._last_read = ""

    @property
    def last_read(self):
        """Get the readonly last read."""
        return self._last_read

    def handle_line(self, data):
        """Store the latest line in last_read."""
        self._last_read = data
        # print (f'cSerialReader: data received: {repr(data)}')

    def connection_lost(self, exc):
        """Notify the user that the connection was lost."""
        print(f"SerialReaderFactory: port {repr(self.transport.serial.port)} closed {repr(exc)}")


class BlobTracker(threading.Thread):
    """
    Tracks color blobs in 3D space.

    all tracking is done in a separate thread, so use this class in context manager

    self.start() - starts tracking thread
    self.close() - stops tracking thread

    self.get_poses() - gets latest tracker poses
    """

    def __init__(
        self, cam_index=0, *, focal_length_px=490, ball_radius_cm=2, color_masks={},
    ):
        """
        Create a blob tracker.

        example usage:
        >>> tracker = BlobTracker(color_masks={
        ...         'tracker1':{
        ...             'h':(98, 10),
        ...             's':(200, 55),
        ...             'v':(250, 32)
        ...         },
        ... })

        >>> with tracker:
        ...     for _ in range(3):
        ...         poses = tracker.get_poses() # gets latest poses for the trackers
        ...         print (poses)
        ...         time.sleep(1/60) # some delay, doesn't affect the tracking
        {'tracker1': {'x': 0, 'y': 0, 'z': 0}}
        {'tracker1': {'x': 0, 'y': 0, 'z': 0}}
        {'tracker1': {'x': 0, 'y': 0, 'z': 0}}

        :param cam_index: index of the camera that will be used to track the blobs
        :param focal_length_px: focal length in pixels
        :param ball_radius_cm: the radius of the ball to be tracked in cm
        :param color_masks: color mask parameters, in opencv hsv color space, for color detection
        """
        super().__init__()
        self._vs = read_updates(cam_index)

        time.sleep(2)

        self.camera_focal_length = focal_length_px
        self.BALL_RADIUS_CM = ball_radius_cm
        self.cam_index = cam_index

        self.last_frame = None
        self.time_of_last_frame = -1

        frame, self.can_track = self._try_get_frame()

        if not self.can_track:
            self._vs.end()
            self._vs = None
            raise RuntimeError('invalid video source')


        self.frame_height, self.frame_width, _ = frame.shape if self.can_track else (0, 0, 0)

        self.markerMasks = color_masks

        self.poses = {}
        self.blobs = {}

        self.kalmanTrainSize = 15

        self.kalmanFilterz: Dict[str, LazyKalman] = {}
        self.kalmanWasInited = {}

        self.kalmanFilterz2: Dict[str, LazyKalman] = {}
        self.kalmanWasInited2 = {}

        for i in self.markerMasks:
            self.poses[i] = {"x": 0, "y": 0, "z": 0}
            self.blobs[i] = None

            self.kalmanFilterz[i] = None
            self.kalmanWasInited[i] = False

            self.kalmanFilterz2[i] = None
            self.kalmanWasInited2[i] = False

        # -------------------threading related------------------------

        self.alive = True
        self._lock = threading.Lock()
        self.daemon = False
        self.pose_que = queue.Queue()

    def _try_get_frame(self):
        self._vs.update()
        try:
            frame = self._vs.frames[str(self.cam_index)][0]
            can_track = True
            if frame is not self.last_frame:
                self.time_of_last_frame = time.time()
        except Exception as e:
            print("_try_get_frame() failed with:", repr(e), self._vs.frames)
            frame = None
            can_track = False
        return frame, can_track

    def stop(self):
        """Stop the blob tracking thread."""
        self.alive = False
        self.can_track = False
        self.join(4)

    def run(self):
        """Run the main blob tracking thread."""
        try:
            _, self.can_track = self._try_get_frame()

            if not self.can_track:
                raise RuntimeError("video source already expired")

        except Exception as e:
            print(f"failed to start thread, video source expired?: {repr(e)}")
            self.alive = False
            return

        while self.alive and self.can_track:
            try:
                self.find_blobs_in_frame()
                self.solve_blob_poses()
                # rotate(self.poses, self.offsets)

                if not self.pose_que.empty():
                    try:
                        self.pose_que.get_nowait()

                    except queue.Empty:
                        pass

                self.pose_que.put(self.poses.copy())

                if not self.can_track:
                    break

            except Exception as e:
                print(f"tracking failed: {repr(e)}")
                break

        self.alive = False
        self.can_track = False

    def get_poses(self) -> Dict[str, Dict[str, float]]:
        """Get the least recent set of poses."""
        if self.can_track and self.alive:
            return self.pose_que.get()
        else:
            raise RuntimeError("tracking loop already finished")

    def find_blobs_in_frame(self):
        """Get a frame from the camera and find all the blobs in it."""
        if self.alive:
            frame, self.can_track = self._try_get_frame()

            for key, mask_range in self.markerMasks.items():
                hc, hr = mask_range["h"]

                sc, sr = mask_range["s"]

                vc, vr = mask_range["v"]

                color_high = [hc + hr, sc + sr, vc + vr]
                color_low = [hc - hr, sc - sr, vc - vr]

                if self.can_track:
                    blurred = cv2.GaussianBlur(frame, (3, 3), 0)

                    hsv = cv2.cvtColor(blurred, cv2.COLOR_BGR2HSV)

                    mask = cv2.inRange(hsv, tuple(color_low), tuple(color_high))

                    temp = cv2.findContours(mask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

                    if len(temp) == 3:
                        _, cnts, hr = temp

                    else:
                        cnts, hr = temp

                    if len(cnts) > 0:
                        c = max(cnts, key=cv2.contourArea)
                        self.blobs[key] = c if len(c) >= 5 and cv2.contourArea(c) > 10 else None

    def solve_blob_poses(self):
        """Solve for and set the poses of all blobs visible by the camera."""
        for key, blob in self.blobs.items():
            if blob is not None:
                elip = cv2.fitEllipse(blob)

                x, y = elip[0]
                w, h = elip[1]

                if not self.kalmanWasInited2[key]:
                    self.kalmanFilterz2[key] = LazyKalman([x, y, w], np.eye(3), np.eye(3))
                    self.kalmanWasInited2[key] = True

                else:
                    x, y, w = self.kalmanFilterz2[key].apply([x, y, w])

                f_px = self.camera_focal_length
                x_px = x
                y_px = y
                a_px = w / 2

                x_px -= self.frame_width / 2
                y_px = self.frame_height / 2 - y_px

                l_px = np.sqrt(x_px ** 2 + y_px ** 2)

                k = l_px / f_px

                j = (l_px + a_px) / f_px

                l = (j - k) / (1 + j * k)

                d_cm = self.BALL_RADIUS_CM * np.sqrt(1 + l ** 2) / l

                fl = f_px / l_px

                z_cm = d_cm * fl / np.sqrt(1 + fl ** 2)

                l_cm = z_cm * k

                x_cm = l_cm * x_px / l_px
                y_cm = l_cm * y_px / l_px

                self.poses[key]["x"] = x_cm / 100
                self.poses[key]["y"] = y_cm / 100
                self.poses[key]["z"] = z_cm / 100

                # self._translate()

                if not has_nan_in_pose(self.poses[key]):
                    if not self.kalmanWasInited[key]:
                        self.kalmanFilterz[key] = LazyKalman(
                            [self.poses[key]["x"], self.poses[key]["y"], self.poses[key]["z"],], np.eye(3), np.eye(3)
                        )
                        self.kalmanWasInited[key] = True
                    else:
                        self.poses[key]["x"], self.poses[key]["y"], self.poses[key]["z"] = self.kalmanFilterz[
                            key
                        ].apply([self.poses[key]["x"], self.poses[key]["y"], self.poses[key]["z"]])

    def close(self):
        """Close the blob tracker thread."""
        with self._lock:
            self.stop()

            if self._vs is not None:
                self._vs.end()
                self._vs = None

    def __enter__(self):
        self.start()
        if not self.alive:
            self.close()
            raise RuntimeError("video source already expired")

        return self

    def __exit__(self, *exc):
        self.close()



'''pose objects'''


class PoseEuler(object):
    """Struct containing all variables needed for instantaneous pose."""

    __slots__ = [
        "x",
        "y",
        "z",
        "vel_x",
        "vel_y",
        "vel_z",
        "acc_x",
        "acc_y",
        "acc_z",
        "ang_x",
        "ang_y",
        "ang_z",
        "vel_ang_x",
        "vel_ang_y",
        "vel_ang_z",
        "acc_ang_x",
        "acc_ang_y",
        "acc_ang_z",
    ]

    def __init__(
        self,
        x: float = 0,
        y: float = 0,
        z: float = 0,
        vel_x: float = 0,
        vel_y: float = 0,
        vel_z: float = 0,
        acc_x: float = 0,
        acc_y: float = 0,
        acc_z: float = 0,
        ang_x: float = 0,
        ang_y: float = 0,
        ang_z: float = 0,
        vel_ang_x: float = 0,
        vel_ang_y: float = 0,
        vel_ang_z: float = 0,
        acc_ang_x: float = 0,
        acc_ang_y: float = 0,
        acc_ang_z: float = 0,
    ):
        """
        Instantaneous pose and velocity of an object.

        :param pose: location in meters and orientation in quaternion
        :param velocity: velocity in meters/second
                         Angular velocity of the pose in axis-angle representation. The direction is the angle of
                         rotation and the magnitude is the angle around that axis in radians/second.
        """
        self.x = x
        self.y = y
        self.z = z
        self.vel_x = vel_x
        self.vel_y = vel_y
        self.vel_z = vel_z
        self.acc_x = acc_x
        self.acc_y = acc_y
        self.acc_z = acc_z
        self.ang_x = ang_x
        self.ang_y = ang_y
        self.ang_z = ang_z
        self.vel_ang_x = vel_ang_x
        self.vel_ang_y = vel_ang_y
        self.vel_ang_z = vel_ang_z
        self.acc_ang_x = acc_ang_x
        self.acc_ang_y = acc_ang_y
        self.acc_ang_z = acc_ang_z

    def __getitem__(self, key):
        if key in self.__slots__:
            return getattr(self, key)

        raise KeyError(key)

    def __setitem__(self, key, val):
        assert isinstance(val, numbers.Number), "value should be numeric"

        if key in self.__slots__:
            setattr(self, key, val)
            return

        raise KeyError(key)

    def __repr__(self):
        return "<{}.{} [{}\n] object at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            "\n\t".join([f"{key}={self[key]}" for key in self.__slots__]),
            hex(id(self)),
        )


class Pose(object):
    """Struct containing all variables needed for instantaneous pose."""

    __slots__ = [
        "x",
        "y",
        "z",
        "r_w",
        "r_x",
        "r_y",
        "r_z",
        "vel_x",
        "vel_y",
        "vel_z",
        "ang_vel_x",
        "ang_vel_y",
        "ang_vel_z",
    ]

    def __init__(self, pose=(0, 0, 0, 1, 0, 0, 0), velocity=(0,) * 6):
        """
        Instantaneous pose and velocity of an object.

        :param pose: location in meters and orientation in quaternion
        :param velocity: velocity in meters/second
                         Angular velocity of the pose in axis-angle representation. The direction is the angle of
                         rotation and the magnitude is the angle around that axis in radians/second.
        """
        self.x: float = pose[0]  # +(x) is right in meters
        self.y: float = pose[1]  # +(y) is up in meters
        self.z: float = pose[2]  # -(z) is forward in meters
        self.r_w: float = pose[3]  # from 0 to 1
        self.r_x: float = pose[4]  # from -1 to 1
        self.r_y: float = pose[5]  # from -1 to 1
        self.r_z: float = pose[6]  # from -1 to 1

        self.vel_x: float = velocity[0]  # +(x) is right in meters/second
        self.vel_y: float = velocity[1]  # +(y) is up in meters/second
        self.vel_z: float = velocity[2]  # -(z) is forward in meters/second
        self.ang_vel_x: float = velocity[3]
        self.ang_vel_y: float = velocity[4]
        self.ang_vel_z: float = velocity[5]

    def __getitem__(self, key):
        if key in self.__slots__:
            return getattr(self, key)

        raise KeyError(key)

    def __setitem__(self, key, val):
        assert isinstance(val, numbers.Number), "value should be numeric"

        if key in self.__slots__:
            setattr(self, key, val)
            return

        raise KeyError(key)

    def __repr__(self):
        return "<{}.{} [{}] object at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            ", ".join([f"{key}={self[key]}" for key in self.__slots__]),
            hex(id(self)),
        )

    def __len__(self):
        return len(self.__slots__)

class ControllerState(Pose):
    """Struct containing all variables needed for instantaneous controller state."""

    __slots__ = [
        "x",
        "y",
        "z",
        "r_w",
        "r_x",
        "r_y",
        "r_z",
        "vel_x",
        "vel_y",
        "vel_z",
        "ang_vel_x",
        "ang_vel_y",
        "ang_vel_z",
        "grip",
        "system",
        "menu",
        "trackpad_click",
        "trigger_value",
        "trackpad_x",
        "trackpad_y",
        "trackpad_touch",
        "trigger_click",
    ]

    def __init__(
        self,
        pose=(0, 0, 0, 1, 0, 0, 0),
        velocity=(0,) * 6,
        grip=0,
        system=0,
        menu=0,
        trackpad_click=0,
        trigger_value=0,
        trackpad_x=0,
        trackpad_y=0,
        trackpad_touch=0,
        trigger_click=0,
    ):
        """
        Instantaneous controller state.

        :param pose: location in meters and orientation in quaternion
        :param velocity: velocity in meters/second
                         Angular velocity of the pose in axis-angle representation. The direction is the angle of
                         rotation and the magnitude is the angle around that axis in radians/second.
        :param grip: is grip down
        :param system: is system button down
        :param menu: is menu button down
        :param trackpad_click: is trackpad clicked
        :param trigger_value: trigger pressure value
        :param trackpad_x: trackpad touch x position
        :param trackpad_y: trackpad touch y position
        :param trackpad_touch: is trackpad being touched
        :param trigger_click: is trigger clicked
        """
        super().__init__(pose, velocity)
        self.grip: int = grip  # 0 or 1
        self.system: int = system  # 0 or 1
        self.menu: int = menu  # 0 or 1
        self.trackpad_click: int = trackpad_click  # 0 or 1
        self.trigger_value: int = trigger_value  # from 0 to 1
        self.trackpad_x: float = trackpad_x  # from -1 to 1
        self.trackpad_y: float = trackpad_y  # from -1 to 1
        self.trackpad_touch: int = trackpad_touch  # 0 or 1
        self.trigger_click: int = trigger_click  # 0 or 1

def get_slot_values(slotted_instance):
    """Get all slot values in a class with slots."""
    # thanks: https://stackoverflow.com/a/6720815/782170
    return [getattr(slotted_instance, slot) for slot in slotted_instance.__slots__]



'''a colection of receivers used in drivers, for debugging, actually used receiver is writen in c++'''

import socket
import time
import threading
from . import utilz as u


class DummyDriverReceiver(threading.Thread):
    """
    no docs, again... too bad!

    example:
    t = DummyDriverReceiver(57, addr='192.168.31.60', port=6969)

    with t:
        t.send('hello') # driver id message
        for _ in range(10):
            time.sleep(1) # delay, doesn't affect the receiving
            print (t.get_pose())
        t.send('nut') # whatever send message

    """


    def __init__(self, expected_pose_size, *, addr="127.0.0.1", port=6969):
        """
        ill let you guess what this does
        """
        super().__init__()
        self.eps = expected_pose_size

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((addr, port))
        self.sock.settimeout(2)

        self.backBuffer = bytearray()
        self.readSize = 50
        self._terminator = b'\n'

        self.newPose = [0 for _ in range(self.eps)]

        # -------------------threading related------------------------

        self.alive = True
        self._lock = threading.Lock()
        self.daemon = False

    def __enter__(self):
        self.start()
        if not self.alive:
            self.close()
            raise RuntimeError("receive thread already finished")

        return self

    def __exit__(self, *exp):
        self.stop()

    def stop(self):
        self.alive = False
        self.join(4)
        time.sleep(0.01)
        self.close()

    def close(self):
        """hammer time!"""
        self.sock.send(b"CLOSE\n")
        time.sleep(1)
        self.sock.close()

    def get_pose(self):
        return self.newPose

    def send(self, text):
        self.sock.send(u.format_str_for_write(text))

    def _handlePacket(self, lastPacket):
        lastPacket = lastPacket.decode('utf-8')
        if self.alive and not u.strings_share_characters(lastPacket.lower(), "qwrtyuiopsasdfghjklzxcvbnm><*[]{}()"):
            pose = u.get_numbers_from_text(lastPacket.strip('\n').strip(' '), " ")

            if len(pose) != self.eps:
                print (f'pose size miss match, expected {self.eps}, got {len(pose)}')
                return False

            else:
                self.newPose = pose
                return True

        return False

    def run(self):
        while self.alive:
            try:
                data = self.sock.recv(self.readSize)
                self.backBuffer.extend(data)

                if not data:
                    break

                while self._terminator in self.backBuffer:
                    lastPacket, self.backBuffer = self.backBuffer.split(self._terminator, 1)
                    if not self._handlePacket(lastPacket):
                        print (repr(lastPacket), repr(self.backBuffer))

            except socket.timeout as e:
                pass

            except Exception as e:
                print(f"DummyDriverReceiver receive thread failed: {repr(e)}")
                break

        self.alive = False


class UduDummyDriverReceiver(threading.Thread):
    """
    no docs, again... too bad!

    example:
    t = UduDummyDriverReceiver('h13 c22 c22')

    with t:
        t.send('hello') # driver id message
        for _ in range(10):
            time.sleep(1) # delay, doesn't affect the receiving
            print (t.get_pose())
        t.send('nut') # whatever send message

    """


    def __init__(self, expected_pose_struct, *, addr="127.0.0.1", port=6969):
        """
        ill let you guess what this does
        """
        super().__init__()
        self.device_order, self.eps = u.get_pose_struct_from_text(expected_pose_struct)

        if not self.eps:
            raise RuntimeError(f'invalid expected_pose_struct: {expected_pose_struct}')

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((addr, port))
        self.sock.settimeout(2)

        self.backBuffer = bytearray()
        self.readSize = 50
        self._terminator = b'\n'

        self.newPoses = [tuple(0 for _ in range(i)) for i in self.eps]

        # -------------------threading related------------------------

        self.alive = True
        self._lock = threading.Lock()
        self.daemon = False

    def __enter__(self):
        self.start()
        if not self.alive:
            self.close()
            raise RuntimeError("receive thread already finished")

        return self

    def __exit__(self, *exp):
        self.stop()

    def stop(self):
        self.alive = False
        self.join(4)
        time.sleep(0.01)
        self.close()

    def close(self):
        """hammer time!"""
        self.sock.send(b"CLOSE\n")
        time.sleep(1)
        self.sock.close()

    def get_pose(self):
        return self.newPoses

    def send(self, text):
        self.sock.send(u.format_str_for_write(text))

    def _handlePacket(self, lastPacket):
        lastPacket = lastPacket.decode('utf-8')
        if self.alive and not u.strings_share_characters(lastPacket.lower(), "qwrtyuiopsasdfghjklzxcvbnm><*[]{}()"):
            poses = u.parse_poses_from_packet(u.get_numbers_from_text(lastPacket.strip('\n').strip(' '), " "), self.eps)


            if u.get_poses_shape(poses) != self.eps:
                print (f'pose size miss match, expected {self.eps}, got {u.get_poses_shape(poses)}')
                return False

            else:
                self.newPoses = poses
                return True

        return False

    def run(self):
        while self.alive:
            try:
                data = self.sock.recv(self.readSize)
                self.backBuffer.extend(data)

                if not data:
                    break

                while self._terminator in self.backBuffer:
                    lastPacket, self.backBuffer = self.backBuffer.split(self._terminator, 1)
                    if not self._handlePacket(lastPacket):
                        print (repr(lastPacket), repr(self.backBuffer))

            except socket.timeout as e:
                pass

            except Exception as e:
                print(f"UduDummyDriverReceiver receive thread failed: {repr(e)}")
                break

        self.alive = False



import pathlib

import cv2
import numpy as np
from PIL import Image

ourpath = pathlib.Path(__file__).parent

menu_img = Image.open(ourpath.joinpath("ai_menu.png").absolute())
menu_np = np.copy(np.asarray(menu_img))
menu_np.setflags(write=True)
red = np.copy(menu_np[:, :, 0])
blue = menu_np[:, :, 2]
menu_np[:, :, 0] = blue
menu_np[:, :, 2] = red



"""
pyvr.

The most commonly used pyvr commands are:
   calibrate  Calibrate the vr headset's tracking
   track      Start the tracking system
   server     Start the server that communicates with SteamVR

See 'pyvr help <command>' for more information on a specific command.

usage: pyvr <command> [<args>...]

options:
   -h, --help
"""
import logging

from docopt import docopt

from virtualreality import __version__


def main():
    logger = logging.getLogger()
    """Pyvr entry point."""
    args = docopt(__doc__, version=f"pyvr version {__version__}", options_first=True)

    logger.debug("command: {}".format(args["<command>"]))

    if args["<command>"] == "calibrate":
        from virtualreality.calibration.manual_color_mask_calibration import main

        main()
    elif args["<command>"] == "calibrate-cam":
        from virtualreality.calibration.camera_calibration import main

        main()
    elif args["<command>"] == "track":
        from virtualreality.trackers.color_tracker import main

        main()
    elif args["<command>"] == "server":
        from virtualreality.server.server import run_til_dead

        run_til_dead()
    elif args["<command>"] in ["help", None]:
        docopt(__doc__, argv="--help")
    else:
        exit(f"{args['<command>']} is not a valid pyvr command. See 'pyvr help'")


if __name__ == "__main__":
    main()



import numpy as np
from pytest import skip
from virtualreality.util.IMU import get_i2c_imu, get_coms_in_range


# @skip("Not an actual test yet")
def test_get_i2c_imu():
    coms = get_coms_in_range()
    con, pro = get_i2c_imu(coms[0])
    pro.start()
    orient1 = (float("nan"),) * 3
    while np.any(np.isnan(orient1)):
        orient1 = pro.protocol.imu.get_instant_orientation()
    while True:
        # print(f"acc: {pro.protocol.imu.get_acc(Quaternion.from_euler(3*pi/2, pi/4, 0))}")
        # pro.protocol.imu.get_grav(stationary=True)
        # print(f"grav:{pro.protocol.imu.grav_magnitude}")
        # print(f"gyro: {pro.protocol.imu.get_gyro()}")
        # print(f"mag: {pro.protocol.imu.get_mag()}")
        print(f"orient1: {orient1}")
        print(f"orient2: {pro.protocol.imu.get_instant_orientation(orient1)}")


if __name__ == "__main__":
    test_get_i2c_imu()



'''base template class'''
import asyncio
import numbers
import warnings

from ..util import utilz as u


class KeepAliveTrigger:
    '''a keep alive trigger, used in poser templates for threading signals'''
    __slots__ = ['is_alive', 'sleep_delay']
    def __init__(self, is_alive, sleep_delay):
        '''
        :is_alive: bool, stored in self.is_alive
        :sleep_delay: sleep delay in seconds, stored in self.sleep_delay
        '''
        self.is_alive = is_alive
        self.sleep_delay = sleep_delay


class PoserTemplateBase(object):
    """
    Base template class, should only be used to create new poser templates


    supplies a last message from server buffer:
        self.last_read - bytearray containing the last message from the server, it is recommended to consume it, it will be populated with new data when its received

    supplies threading vars:
        self.coro_list - list of all methods recognized as threads
        self.coro_keep_alive - dict of all registered threads, containing self.coro_keepAlive['memberThreadName'] = KeepAliveTrigger(is_alive, sleep_delay), this dict is populated at self.main() call

    this class also has 3 built in threads, it is not recommended you override any of them:
        self.send - sends all pose data to the server
        self.recv - receives messages from the server, the last message is stored in self.lastRead
        self.close - closes the connection to the server and ends all threads that where registered by register_member_thread when the 'q' key is pressed

    this class will assume that every
    child method without '_' as a first character in the name is a thread,
    unless the name of that method has been added to self._coro_name_exceptions

    every child class also needs to register it's thread
    methods with the PoserTemplate.register_member_thread decorator
    """

    def __init__(
        self, *, addr="127.0.0.1", port=6969, send_delay=1 / 100, recv_delay=1 / 1000, **kwargs,
    ):
        """
        :addr: is the address of the server to connect to, stored in self.addr
        :port: is the port of the server to connect to, stored in self.port
        :send_delay: sleep delay for the self.send thread(in seconds)
        :recv_delay: sleep delay for the self.recv thread(in seconds)
        """
        self.addr = addr
        self.port = port

        self._coro_name_exceptions = ["main", "register_member_thread"]

        self.coro_list = [
            method_name
            for method_name in dir(self)
            if callable(getattr(self, method_name))
            and method_name[0] != "_"
            and method_name not in self._coro_name_exceptions
        ]

        self.coro_keep_alive = {
            "close": KeepAliveTrigger(True, 0.1),
            "send": KeepAliveTrigger(True, send_delay),
            "recv": KeepAliveTrigger(True, recv_delay),
        }
        self.last_read = ""
        self.id_message = 'holla'

    async def _socket_init(self):
        """
        Connect to the server using self.addr and self.port and send the id message.

        Also store socket reader and writer in self.reader and self.writer.
        It is not recommended you override this method
        """
        print(f'connecting to the server at "{self.addr}:{self.port}"...')
        # connect to the server
        self.reader, self.writer = await asyncio.open_connection(self.addr, self.port, loop=asyncio.get_event_loop())
        # send poser id message
        self.writer.write(u.format_str_for_write(self.id_message))

    async def send(self):
        """Send all poses thread, you need to implement this!"""
        raise NotImplementedError('please implement the send thread')

    async def recv(self):
        """Receive messages thread."""
        while self.coro_keep_alive["recv"].is_alive:
            try:
                data = await u.read3(self.reader)
                self.last_read = data

                await asyncio.sleep(self.coro_keep_alive["recv"].sleep_delay)
            except Exception as e:
                print(f"recv failed: {e}")
                self.coro_keep_alive["recv"].is_alive = False
                break

    async def close(self):
        """Await close thread, press "q" to kill all registered threads."""
        try:
            import keyboard

            while self.coro_keep_alive["close"].is_alive:
                if keyboard.is_pressed("q"):
                    break

                await asyncio.sleep(self.coro_keep_alive["close"].sleep_delay)

        except ImportError as e:
            print(f"close: failed to import keyboard, poser will close in 10 seconds: {e}")
            await asyncio.sleep(10)

        print("closing...")
        self.coro_keep_alive["close"].is_alive = False
        for key in self.coro_keep_alive:
            if self.coro_keep_alive[key].is_alive:
                self.coro_keep_alive[key].is_alive = False
                print(f"{key} stop sent...")
            else:
                print(f"{key} already stopped")

        await asyncio.sleep(0.5)

        try:
            self.writer.write(u.format_str_for_write("CLOSE"))
            self.writer.close()
            await self.writer.wait_closed()

        except Exception as e:
            print (f'failed to close connection: {e}')

        print("finished")

    async def main(self):
        """
        Run the main async method.

        Gathers all recognized threads and runs them.
        It is not recommended you override this method.
        """
        await self._socket_init()

        await asyncio.gather(
            *[getattr(self, coro_name)() for coro_name in self.coro_list if coro_name not in self._coro_name_exceptions]
        )

    @staticmethod
    def register_member_thread(sleepDelay, runInDefaultExecutor=False):
        """
        Registers thread functions, should only be used on member functions

        sleepDelay - sleep delay in seconds
        runInDefaultExecutor - bool, set True if you want the function to be executed in asyncio's default pool executor
        """

        def _thread_reg(func):
            def _thread_reg_wrapper(self, *args, **kwargs):

                if not asyncio.iscoroutinefunction(func) and not runInDefaultExecutor:
                    raise ValueError(f"{repr(func)} is not a coroutine function and runInDefaultExecutor is set to False")

                if func.__name__ not in self.coro_keep_alive and func.__name__ in self.coro_list and func.__name__ not in self._coro_name_exceptions:
                    self.coro_keep_alive[func.__name__] = KeepAliveTrigger(True, sleepDelay)

                elif func.__name__ in self._coro_name_exceptions:
                    warnings.warn("thread register ignored, thread name is in balck list", RuntimeWarning)
                    return func(self, *args, **kwargs)

                else:
                    raise NameError(
                        f"trying to register existing thread, thread with name {repr(func.__name__)} already exists"
                    )

                if runInDefaultExecutor:
                    loop = asyncio.get_running_loop()

                    return loop.run_in_executor(None, func, self, *args, **kwargs)

                return func(self, *args, **kwargs)

            setattr(_thread_reg_wrapper, "__name__", func.__name__)

            return _thread_reg_wrapper

        return _thread_reg

class PoserClientBase(PoserTemplateBase):
    '''
    poser client base class
    should only be inherited together with one of the poser templates
    should only be used to crate new poser clients
    '''
    def __init__(self, *args, **kwargs):
        """init"""
        super().__init__(*args, **kwargs)
        self._coro_name_exceptions.append("thread_register")

    def thread_register(self, sleep_delay, runInDefaultExecutor=False):
        """
        Registers a thread for the client

        sleepDelay - sleep delay in seconds
        runInDefaultExecutor - bool, set True if you want the function to be executed in asyncio's default pool executor
        """

        def _thread_register(coro):
            if not asyncio.iscoroutinefunction(coro) and not runInDefaultExecutor:
                raise ValueError(f"{repr(coro)} is not a coroutine function and runInDefaultExecutor is set to False")

            if coro.__name__ not in self.coro_keep_alive and coro.__name__ not in self.coro_list:
                self.coro_keep_alive[coro.__name__] = KeepAliveTrigger(True, sleep_delay)
                self.coro_list.append(coro.__name__)

                if runInDefaultExecutor:

                    def _wrapper(*args, **kwargs):
                        setattr(_wrapper, "__name__", f"{coro.__name__} _decorated")
                        loop = asyncio.get_running_loop()

                        return loop.run_in_executor(None, coro, *args, **kwargs)

                    setattr(_wrapper, "__name__", coro.__name__)
                    setattr(self, coro.__name__, _wrapper)
                    return _wrapper

                else:
                    setattr(self, coro.__name__, coro)

            else:
                raise NameError(
                    f"trying to register existing thread, thread with name {repr(coro.__name__)} already exists"
                )

            return coro

        return _thread_register



import multiprocessing
import pathlib
import sys
from typing import Union

import librosa
import numpy as np
import pyaudio

ourpath = pathlib.Path(__file__).parent


class InputConfig(object):
    min_seconds = 2
    sampling_rate = 44100
    stream_chunk_size = 1024
    input_channels = 1


easy_music = next(ourpath.joinpath('easy_music').glob('*.mp3'))
easy_bpm = int(open(str(next(ourpath.joinpath('easy_music').glob('*.txt')))).read())
normal_music = next(ourpath.joinpath('normal_music').glob('*.mp3'))
normal_bpm = int(open(str(next(ourpath.joinpath('normal_music').glob('*.txt')))).read())
hard_music = next(ourpath.joinpath('hard_music').glob('*.mp3'))
hard_bpm = int(open(str(next(ourpath.joinpath('hard_music').glob('*.txt')))).read())
menu_music = next(ourpath.joinpath('menu_music').glob('*.mp3'))
menu_bpm = int(open(str(next(ourpath.joinpath('menu_music').glob('*.txt')))).read())


def read_audio_file(file_path: Union[str, pathlib.Path], config: InputConfig):
    """
    Reads in an audio file and converts to a numpy array.
    >>> f = read_audio_file('./sw.mp3', InputConfig())
    >>> print(f)
    [[-1.2063791e-04 -1.0168436e-04 -3.2342479e-05 ... -7.7938515e-05
      -8.8288434e-05 -5.1688534e-05]
     [ 9.4297948e-06 -2.1362750e-05 -7.0384616e-05 ...  3.9314014e-05
       5.7227458e-05  3.8280330e-05]]

    """
    if not isinstance(file_path, str):
        file_path = str(file_path)
    min_samples = int(config.min_seconds * config.sampling_rate)
    try:
        audio_time_series, sampling_rate = librosa.load(
            file_path, sr=config.sampling_rate, mono=False
        )
        trimmed_audio, trim_indexes = librosa.effects.trim(audio_time_series)

        if trimmed_audio.shape[-1] < min_samples:
            padding = min_samples - len(trimmed_audio)
            offset = padding // 2
            trimmed_audio = np.pad(
                trimmed_audio, (offset, padding - offset), "constant"
            )
        return trimmed_audio
    except Exception as e:
        print(f"Exception while reading {file_path}: {e}", sys.stderr)
        return np.zeros(min_samples, dtype=np.float32)


def play_audio(array: np.ndarray, config: InputConfig, loop=False):
    """
    Plays a numpy array as audio.

    from: https://www.scivision.dev/playing-sounds-from-numpy-arrays-in-python/

    >>> i = InputConfig()
    >>> f = read_audio_file('./sw.mp3', i)
    >>> play_audio(f, i)

    # song: https://www.youtube.com/watch?v=d-nxW9qBtxQ

    """
    if len(array.shape) == 2:
        channels = int(array.shape[0])
    else:
        channels = 1

    array = np.transpose(array)
    p = pyaudio.PyAudio()
    stream = p.open(rate=config.sampling_rate, format=pyaudio.paFloat32, channels=channels, output=True)
    arrby = array.tobytes()
    stream.write(arrby)
    while loop:
        stream.write(arrby)
    stream.close()
    p.terminate()


def get_audio_thread(music: Union[str, pathlib.Path], loop=False):
    audio_cfg = InputConfig()
    np_file = read_audio_file(music, audio_cfg)
    audio_thread = multiprocessing.Process(target=play_audio, args=(np_file, audio_cfg, loop))
    return audio_thread



from unittest import mock

import numpy as np
from tests.data import rgby_tracking_image_np
from tests.mocks.mock_displayarray import MockReadUpdates
from virtualreality.util.utilz import BlobTracker


def test_init():
    with MockReadUpdates(0, [rgby_tracking_image_np]):
        with mock.patch("virtualreality.utilz.time.sleep"):
            b = BlobTracker(color_masks={"tracker1": {"h": (98, 10), "s": (200, 55), "v": (250, 32)}})

            assert b.camera_focal_length == 490
            assert b.BALL_RADIUS_CM == 2
            assert b.offsets == [0, 0, 0]
            assert b.cam_index == 0

            assert b.can_track

            assert b.frame_height == 358
            assert b.frame_width == 638

            assert b.markerMasks == {"tracker1": {"h": (98, 10), "s": (200, 55), "v": (250, 32)}}

            assert b.poses == {"tracker1": {"x": 0, "y": 0, "z": 0}}
            assert b.blobs == {"tracker1": None}


def test_find_blobs_in_frame():
    with MockReadUpdates(0, [rgby_tracking_image_np]):
        with mock.patch("virtualreality.utilz.time.sleep"):
            b = BlobTracker(color_masks={"tracker1": {"h": (98, 10), "s": (200, 55), "v": (250, 32)}})

            b.find_blobs_in_frame()

            assert (
                b.blobs["tracker1"]
                == np.asarray(
                    [
                        [[196, 272]],
                        [[194, 274]],
                        [[193, 274]],
                        [[192, 275]],
                        [[192, 276]],
                        [[191, 277]],
                        [[191, 282]],
                        [[192, 283]],
                        [[192, 284]],
                        [[195, 287]],
                        [[197, 287]],
                        [[198, 288]],
                        [[199, 288]],
                        [[200, 287]],
                        [[203, 287]],
                        [[205, 285]],
                        [[205, 284]],
                        [[206, 283]],
                        [[206, 278]],
                        [[205, 277]],
                        [[205, 276]],
                        [[203, 274]],
                        [[201, 274]],
                        [[200, 273]],
                        [[197, 273]],
                    ],
                    dtype=np.int32,
                )
            ).all()


def test_solve_blob_poses():
    with MockReadUpdates(0, [rgby_tracking_image_np]):
        with mock.patch("virtualreality.utilz.time.sleep"):
            b = BlobTracker(color_masks={"tracker1": {"h": (98, 10), "s": (200, 55), "v": (250, 32)}})

            b.find_blobs_in_frame()
            b.solve_blob_poses()

            assert b.poses == {
                "tracker1": {"x": -0.34523609690742246, "y": -0.28970498418772483, "z": 1.4035486226748464}
            }

            b.solve_blob_poses()

            assert b.poses == {"tracker1": {"x": -0.6575314351920128, "y": 0.11647778625690625, "z": 1.466077432585446}}



from unittest import mock

from displayarray import read_updates
from tests.data import rgby_tracking_image_np
from tests.mocks.mock_displayarray import MockReadUpdates


def test_mocks_read():
    with MockReadUpdates(0, [rgby_tracking_image_np]):
        a = read_updates()
        assert isinstance(a, mock.MagicMock)
        a.update()
        assert (a.frames["0"][0] == rgby_tracking_image_np).all()



"""Headset and controller tracers."""
from ..logging import log

logger = log.setup_custom_logger(name=__name__,
    level="INFO",
    file="../logs/trackers.log")

logger.debug("created trackers.log log file")


"""Server loop that communicates between the driver and posers."""
import asyncio

from ..templates import PoserTemplate
from ..util import utilz as u

DOMAIN = (None, 6969)

class Server:
    def __init__(self):
        '''
        im not sure about this on chief,
        looks like a destructor to me
        '''
        self.conz = []
        self.debug = False

        self._driver_idz = [b'hello',]
        self._poser_idz = [b'holla',]
        self._terminator = b'\n'
        self._close_msg = b'CLOSE\n'

    def __repr__(self):
        '''do i need to explain this?'''
        return f'<{self.__class__.__module__}.{self.__class__.__name__} debug={self.debug} active_conz={len(self.conz)} object at {hex(id(self))}>'

    async def send_to_all(self, msg, me):
        '''send a message to all registered connections that are not self and have the corresponding id'''
        try:
            for i in self.conz:
                if i != me and (i[2] == me[2] or i[3]):
                    i[1].write(msg)
                    await i[1].drain()

        except Exception as e:
            print (f'message from {me[0]} lost: {e}')


    async def __call__(self, reader, writer):
        '''this is will run for each incoming connection'''

        addr = writer.get_extra_info("peername")

        id_msg, first_msg = (await reader.read(50)).split(self._terminator)

        me = (addr, writer, id_msg in self._driver_idz or id_msg in self._poser_idz, id_msg in self._poser_idz, f'lol{len(self.conz)}')
        if me not in self.conz:
            print (f'new connection from {addr}')
            self.conz.append(me)

        if first_msg:
            await self.send_to_all(first_msg, me)

        # this is does nothing but looks pretty
        if id_msg in self._driver_idz:
            print ('its a driver')

        elif id_msg in self._poser_idz:
            print ('its a poser')

        # main receive/transmit loop
        while 1:
            try:
                data = await reader.read(400)

                if not data or self._close_msg in data:
                    break

                await self.send_to_all(data, me)

                if self.debug:
                    print (f'{repr(data)} from {addr}')

            except Exception as e:
                print (f'{addr} broke: {e}')
                break

        if me in self.conz:
            self.conz.remove(me)

        try:
            writer.close()
            await writer.wait_closed()
        except Exception as e:
            print (f'error on {addr} close: {e}')

        print (f'connection to {addr} closed')

def run_til_dead(poser: PoserTemplate = None, conn_handle=Server()):
    """Run the server until it dies."""
    loop = asyncio.get_event_loop()
    coro = asyncio.start_server(conn_handle, *DOMAIN, loop=loop)
    server = loop.run_until_complete(coro)

    if poser is not None:
        poser_result = asyncio.run_coroutine_threadsafe(poser.main(), loop)

    # Serve requests until Ctrl+C is pressed
    print("Serving on {}".format(server.sockets[0].getsockname()))
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    # Close the server
    server.close()
    loop.run_until_complete(server.wait_closed())
    loop.close()



import time

import numpy as np
import serial
import serial.threaded
from pyrr import Quaternion
from serial import SerialException

from virtualreality.util import utilz as u


def unit_vector(vector):
    """ Returns the unit vector of the vector.  """
    # https://stackoverflow.com/a/13849249/782170
    vector = np.asarray(vector)
    norm = np.linalg.norm(vector)
    if norm != 0:
        return vector / norm
    else:
        return vector


def angle_between(v1, v2):
    """ Returns the angle in radians between vectors 'v1' and 'v2'::

            >>> angle_between((1, 0, 0), (0, 1, 0))
            1.5707963267948966
            >>> angle_between((1, 0, 0), (1, 0, 0))
            0.0
            >>> angle_between((1, 0, 0), (-1, 0, 0))
            3.141592653589793
    """
    # https://stackoverflow.com/a/13849249/782170
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    cos_theta = np.clip(np.dot(v1_u, v2_u), -1.0, 1.0)
    if cos_theta > 0:
        return np.arccos(cos_theta)
    else:
        return np.arccos(cos_theta) - np.pi / 2.0


def perpendicular_vector(v):
    r""" Finds an arbitrary perpendicular vector to *v*."""
    # https://codereview.stackexchange.com/a/43937
    # for two vectors (x, y, z) and (a, b, c) to be perpendicular,
    # the following equation has to be fulfilled
    #     0 = ax + by + cz

    # x = y = z = 0 is not an acceptable solution
    if v.x == v.y == v.z == 0:
        raise ValueError("zero-vector")

    # If one dimension is zero, this can be solved by setting that to
    # non-zero and the others to zero. Example: (4, 2, 0) lies in the
    # x-y-Plane, so (0, 0, 1) is orthogonal to the plane.
    if v.x == 0:
        return np.asarray([1, 0, 0])
    if v.y == 0:
        return np.asarray([0, 1, 0])
    if v.z == 0:
        return np.asarray([0, 0, 1])

    # arbitrarily set a = b = 1
    # then the equation simplifies to
    #     c = -(x + y)/z
    return np.asarray([1, 1, -1.0 * (v.x + v.y) / v.z])


def from_to_quaternion(v1, v2, is_unit=True, epsilon=1e-5):
    # https://stackoverflow.com/a/11741520/782170
    if not is_unit:
        v1 = unit_vector(v1)
        v2 = unit_vector(v2)

    k_cos_theta = np.dot(v1, v2)
    k = np.sqrt(np.dot(v1, v1) * np.dot(v2, v2))

    if abs(k_cos_theta / k + 1) < epsilon:
        # 180 degree rotation around any orthogonal vector
        return Quaternion(0, unit_vector(perpendicular_vector(v1)))

    q = Quaternion(k_cos_theta + k, *np.cross(v1, v2))
    return q.normalize


def get_ki_kbias(convergance_time, overshoot_measure, sampling_rate):
    ki = sampling_rate / (1.4 * convergance_time + sampling_rate)
    kbias = (overshoot_measure ** 2 * ki) / (160 * convergance_time)
    return ki, kbias


class IMU(object):
    epsilon = 1e-5

    def __init__(self):
        self._acc_x = 0.0
        self._acc_y = 0.0
        self._acc_z = 0.0
        self._gyro_x = 0.0
        self._gyro_y = 0.0
        self._gyro_z = 0.0
        self._mag_x = 0.0
        self._mag_y = 0.0
        self._mag_z = 0.0
        self.grav_magnitude = 14  # I have no idea why gravity is 14 on my imu, but it is
        self._mag_iron_offset_x = 23.5
        self._mag_iron_offset_y = 23.5
        self._mag_iron_offset_z = -45
        self.prev_up = None
        self.prev_north = None
        self.prev_west = None
        self.prev_orientation_time = None
        self.prev_bias = None

    def get_north_west_up(self):
        grav = np.asarray(unit_vector(self.get_grav()))
        mag = unit_vector(np.asarray(self.get_mag()))
        east = unit_vector(np.cross(mag, grav))
        north = -unit_vector(np.cross(east, grav))
        up = -unit_vector(np.cross(north, east))

        return north, -east, up

    def get_instant_orientation(self):
        """Gets orientation quaternion"""
        north, west, up = self.get_north_west_up()
        rot = np.asarray([north, west, up]).transpose()
        q = Quaternion.from_matrix(rot)
        return q

    def get_orientation(self, convergence_acc: float, overshoot_acc: float, convergence_mag: float,
                        overshoot_mag: float):
        # https://www.sciencedirect.com/science/article/pii/S2405896317321201
        if self.prev_up is not None:
            t2 = time.time()

            k_acc, k_bias_acc = get_ki_kbias(convergence_acc, overshoot_acc, t2 - self.prev_orientation_time)
            k_mag, k_bias_mag = get_ki_kbias(convergence_mag, overshoot_mag, t2 - self.prev_orientation_time)

            grav = np.asarray(unit_vector(self.get_grav()))
            mag = unit_vector(np.asarray(self.get_mag()))
            east = unit_vector(np.cross(mag, grav))
            west = -east
            north = -unit_vector(np.cross(east, grav))
            up = -unit_vector(np.cross(north, east))

            gyro = -self.get_gyro()
            if self.prev_bias is not None:
                if np.isnan(self.prev_bias).any():
                    self.prev_bias = None
                else:
                    gyro += self.prev_bias

            gy = Quaternion.from_eulers(gyro * (t2 - self.prev_orientation_time))
            pred_west = (~gy * Quaternion(np.pad(self.prev_west, (0,1), 'constant', constant_values=0)) * gy)
            pred_up = (~gy * Quaternion(np.pad(self.prev_up, (0,1), 'constant', constant_values=0)) * gy)
            pred_north = (~gy * Quaternion(np.pad(self.prev_north, (0,1), 'constant', constant_values=0)) * gy)

            pred_north_v = np.asarray(pred_north.xyz)
            pred_west_v = np.asarray(pred_west.xyz)
            pred_up_v = np.asarray(pred_up.xyz)

            fix_north = Quaternion.from_axis(np.cross(north, pred_north_v) * k_mag)
            fix_west = Quaternion.from_axis(np.cross(west, pred_west_v) * k_mag)
            fix_up = Quaternion.from_axis(np.cross(up, pred_west_v) * k_acc)

            x_acc = np.cross(up, pred_up_v)
            x_mag = np.cross(north, pred_north_v)
            pb0 = k_bias_acc*x_acc + k_bias_mag*x_mag
            if self.prev_bias is not None:
                self.prev_bias = self.prev_bias * (t2 - self.prev_orientation_time) + pb0
            else:
                self.prev_bias = pb0

            true_north = unit_vector((pred_north * fix_north).xyz)
            true_up = unit_vector((pred_up * fix_up).xyz)
            true_west = unit_vector((pred_west * fix_west).xyz)

            rot = np.asarray([true_north, true_west, true_up]).transpose()
            q = Quaternion.from_matrix(rot)

            if not np.isnan(north).any():
                self.prev_north = north
            if not np.isnan(west).any():
                self.prev_west = west
            if not np.isnan(up).any():
                self.prev_up = up

            self.prev_orientation_time = t2

            return q
        else:
            grav = np.asarray(unit_vector(self.get_grav()))
            mag = unit_vector(np.asarray(self.get_mag()))
            east = unit_vector(np.cross(mag, grav))
            north = -unit_vector(np.cross(east, grav))
            up = -unit_vector(np.cross(north, east))

            rot = np.asarray([north, -east, up]).transpose()
            q = Quaternion.from_matrix(rot)
            if (not np.isnan(grav).any()) and (not np.isnan(mag).any()):
                self.prev_west = -east
                self.prev_north = north
                self.prev_up = up
                self.prev_orientation_time = time.time()
            return q

    def get_grav(self, q=None, magnitude=None, stationary=False, accel=None):
        if stationary or q is None or ((accel is not None) and (not np.any(accel))):
            g = np.asarray([self._acc_x, self._acc_y, self._acc_z])
            self.grav_magnitude = sum(abs(np.asarray(g)))
            return g
        elif q is not None:
            if not isinstance(q, Quaternion):
                q = Quaternion.from_euler(*q)
            if magnitude is None:
                magnitude = self.grav_magnitude
            # http://www.varesano.net/blog/fabio/simple-gravity-compensation-9-dom-imus
            g = np.asarray([0.0, 0.0, 0.0])

            # get expected direction of gravity
            g[0] = q.x * magnitude
            g[1] = q.y * magnitude
            g[2] = q.z * magnitude

            return g
        elif accel is not None:
            g = np.asarray([self._acc_x, self._acc_y, self._acc_z])
            g -= accel
            self.grav_magnitude = sum(abs(np.asarray(g)))
            return g

    def get_acc(self, q=None):
        g = self.get_grav(q)
        # compensate accelerometer readings with the expected gravity
        return np.asarray([self._acc_x - g[0], self._acc_y - g[1], self._acc_z - g[2]])

    def get_gyro(self):
        return np.asarray([self._gyro_x, self._gyro_y, self._gyro_z])

    def set_mag_iron_offset(self, offset):
        assert len(offset) == 3
        self._mag_iron_offset_x = offset[0]
        self._mag_iron_offset_x = offset[1]
        self._mag_iron_offset_x = offset[2]

    def get_mag(self):
        mag = [
            self._mag_x - self._mag_iron_offset_x,
            self._mag_y - self._mag_iron_offset_y,
            self._mag_z - self._mag_iron_offset_z,
        ]
        return np.asarray(mag)


class PureIMUProtocol(serial.threaded.Packetizer):
    TERMINATOR = b"\n"

    def __init__(self):
        super().__init__()
        self._last_data = None
        self.time_of_last_data = 0
        self.imu = IMU()

    def handle_packet(self, packet):
        pkt_data = u.get_numbers_from_text(packet)
        if len(pkt_data) >= 9:
            self._last_data = pkt_data[:9]
            self.imu._acc_x = self._last_data[0]
            self.imu._acc_y = self._last_data[1]
            self.imu._acc_z = self._last_data[2]
            self.imu._gyro_x = self._last_data[3]
            self.imu._gyro_y = self._last_data[4]
            self.imu._gyro_z = self._last_data[5]
            self.imu._mag_x = self._last_data[6]
            self.imu._mag_y = self._last_data[7]
            self.imu._mag_z = self._last_data[8]
            self.imu.time_of_last_data = time.time()


def get_coms_in_range(start=0, stop=20):
    coms = []
    for i in range(start, stop):
        try:
            arduino = serial.Serial(f"COM{i}", 115200, timeout=0.1)
            arduino.close()
            coms.append(f"COM{i}")
            break
        except SerialException as se:
            if "FileNotFoundError" in str(se):
                continue
            else:
                print(f"COM{i} blocked. Did you try closing Cura?")
                continue
    return coms


def get_i2c_imu(i2c_port="COM0"):
    try:
        connection = serial.Serial(i2c_port, 115200, timeout=1 / 4)
        p = serial.threaded.ReaderThread(connection, PureIMUProtocol)
        return connection, p
    except Exception as e:
        raise RuntimeError(f"Could not open I2C: {e}")



import math as m
import pathlib
import pickle
import time
from typing import Optional, Union

import cv2
import numpy as np
from displayarray import read_updates
from displayarray.input import mouse_loop
from displayarray.window import SubscriberWindows
from displayarray.effects.overlay import overlay_transparent
from displayarray.effects.transform import transform_about_center
from virtualreality.calibration.ai_tracker_training.audio import get_audio_thread, menu_music, easy_music, normal_music, \
    hard_music, easy_bpm, normal_bpm, hard_bpm
from virtualreality.calibration.ai_tracker_training.video import menu_np
from virtualreality.util.IMU import get_coms_in_range, get_i2c_imu, IMU

ourpath = pathlib.Path(__file__).parent


def get_heartbeat_callback(bpm, sharpness=10):
    def heartbeat_callback(t):
        tempo = bpm / 60.0
        val = m.cos(m.pi * t * tempo) ** (2 * sharpness)
        return val

    return heartbeat_callback


class GameState():
    def __init__(self, cam, imu, out_name, music_file=None, loop=False):
        self.cam = cam
        self.imu = imu
        self.out_name = out_name
        self.music_file = music_file
        self.next_state = None
        self.video_shape = None
        self.mouse_loop = mouse_loop(self.on_mouse)
        self.loop = loop

    def run(self) -> Optional['GameState']:
        audio_thread = None
        if self.music_file is not None:
            audio_thread = get_audio_thread(self.music_file, self.loop)
        if not isinstance(self.cam, SubscriberWindows):
            self.cam = read_updates(self.cam, size=(-1, -1), )
        started = False
        f_prev = None
        for f in self.cam:
            if f and f_prev is not next(iter(f.values()))[0]:
                vid: np.ndarray = next(iter(f.values()))[0]
                if not started:
                    if audio_thread is not None:
                        audio_thread.start()
                    self.on_start()
                    started = True
                if self.next_state is not None:
                    audio_thread.terminate()
                    return self.next_state
                self.video_shape = vid.shape
                vid = self.preprocess_video(vid)
                a = [self.on_video(vid)]
                self.cam.display_frames(a)
                f_prev = vid
                audio_thread.join(0)
                if not audio_thread.is_alive():
                    self.next_state = self.on_end()
                    audio_thread.terminate()
                    return self.next_state
        audio_thread.terminate()

    def on_mouse(self, mouse_state):
        pass

    def preprocess_video(self, video: np.ndarray) -> np.ndarray:
        return np.flip(np.copy(video), 1)

    def on_video(self, video: np.ndarray) -> np.ndarray:
        return video

    def on_start(self):
        pass

    def on_end(self) -> Optional['GameState']:
        return MenuState(self.cam, self.imu, self.out_name)


class AIRecordingState(GameState):
    class GoalPose(object):
        x: float = 0
        y: float = 0
        z: float = 1
        roll: float = 0
        pitch: float = 0
        yaw: float = 0

        def __eq__(self, other):
            if isinstance(other, AIRecordingState.GoalPose):
                out = (
                        self.x == other.x and
                        self.y == other.y and
                        self.z == other.z and
                        self.roll == other.roll and
                        self.pitch == other.pitch and
                        self.yaw == other.yaw
                )
                return out

        def __mul__(self, other):
            out_pose = AIRecordingState.GoalPose()
            if isinstance(other, AIRecordingState.GoalPose):
                out_pose.x = self.x * other.x
                out_pose.y = self.y * other.y
                out_pose.z = self.z * other.z
                out_pose.roll = self.roll * other.roll
                out_pose.pitch = self.pitch * other.pitch
                out_pose.yaw = self.yaw * other.yaw
            else:
                out_pose.x = self.x * other
                out_pose.y = self.y * other
                out_pose.z = self.z * other
                out_pose.roll = self.roll * other
                out_pose.pitch = self.pitch * other
                out_pose.yaw = self.yaw * other
            return out_pose

        def __add__(self, other):
            out_pose = AIRecordingState.GoalPose()
            if isinstance(other, AIRecordingState.GoalPose):
                out_pose.x = self.x + other.x
                out_pose.y = self.y + other.y
                out_pose.z = self.z + other.z
                out_pose.roll = self.roll + other.roll
                out_pose.pitch = self.pitch + other.pitch
                out_pose.yaw = self.yaw + other.yaw
            else:
                out_pose.x = self.x + other
                out_pose.y = self.y + other
                out_pose.z = self.z + other
                out_pose.roll = self.roll + other
                out_pose.pitch = self.pitch + other
                out_pose.yaw = self.yaw + other
            return out_pose

    class Record(list):
        def __init__(self, grabbed_frame):
            super().__init__()
            self.grabbed_frame = grabbed_frame

    def __init__(self, cam, grabbed_frame, imu, out_name, music, bpm):
        super().__init__(cam, imu, out_name, music)
        self.heartbeat = get_heartbeat_callback(bpm)
        self.bpm = bpm / 2.0
        self.grabbed_frame = grabbed_frame
        self.t0 = None
        self.prev_beat_distance = 0
        self.goal_pose = AIRecordingState.GoalPose()
        self.prev_goal = AIRecordingState.GoalPose()
        hsv = cv2.cvtColor(self.grabbed_frame, cv2.COLOR_BGR2HSV)
        self.grabbed_frame = np.append(self.grabbed_frame, hsv[:, :, 2:], -1)
        self.goal_square = np.zeros((100, 200, 4))
        self.goal_square[:, :, 3] = 255
        self.goal_square[:, :, 1] = 255
        self.goal_square = overlay_transparent(np.zeros((self.grabbed_frame.shape[0], self.grabbed_frame.shape[1], 3)),
                                               self.goal_square)
        hsv_square = cv2.cvtColor(self.goal_square.astype(np.uint8), cv2.COLOR_BGR2HSV)
        self.goal_square = np.append(self.goal_square, hsv_square[:, :, 2:] * .5, -1)
        self.out_video = cv2.VideoWriter(
            f"{out_name}.avi",
            cv2.VideoWriter_fourcc(*'MJPG'),
            20.0,
            (grabbed_frame.shape[1], grabbed_frame.shape[0]),
            True
        )
        self.out_record = AIRecordingState.Record(self.grabbed_frame)
        self.out_name = out_name

    def on_start(self):
        self.t0 = time.time()

    def make_new_goal(self):
        g = AIRecordingState.GoalPose()
        g.x = (np.random.random_sample() - .5) * self.video_shape[0]
        g.y = (np.random.random_sample() - .5) * self.video_shape[1]
        g.z = (np.random.random_sample() + .25) * 2
        g.roll = (np.random.random_sample() - .5) * 90
        g.yaw = (np.random.random_sample() - .5) * 120
        g.pitch = (np.random.random_sample() - .5) * 120
        self.goal_pose = g

    def preprocess_video(self, video: np.ndarray) -> np.ndarray:
        self.out_video.write(video)
        return super().preprocess_video(video)

    def on_video(self, video: np.ndarray) -> np.ndarray:
        current_time = time.time()
        wall_time = current_time - self.t0
        spb = 60.0 / self.bpm
        wall_beat_distance = (wall_time % spb) / spb
        beat = self.heartbeat(wall_time) / 2.0
        a = (1 - beat) * video + beat * 255
        if wall_beat_distance < self.prev_beat_distance:
            zero_goal = AIRecordingState.GoalPose()
            self.prev_goal = self.goal_pose
            if self.prev_goal == zero_goal:
                self.make_new_goal()
            else:
                self.goal_pose = zero_goal
        beat_distance = (np.tanh(wall_beat_distance * np.pi * 2.0 - np.pi) / 2.0) + .5
        g = self.prev_goal * (1 - beat_distance) + self.goal_pose * beat_distance
        xfromed_grab = transform_about_center(self.grabbed_frame,
                                              scale_multiplier=(g.z, g.z),
                                              translation=(g.x, g.y),
                                              rotation_degrees=g.roll,
                                              skew=(g.yaw, g.pitch))
        xformed_goal = transform_about_center(self.goal_square,
                                              scale_multiplier=(self.goal_pose.z, self.goal_pose.z),
                                              translation=(self.goal_pose.x, self.goal_pose.y),
                                              rotation_degrees=self.goal_pose.roll,
                                              skew=(self.goal_pose.yaw, self.goal_pose.pitch))
        a = overlay_transparent(a, xformed_goal)
        a = overlay_transparent(a, xfromed_grab)
        self.prev_beat_distance = wall_beat_distance
        q = self.imu.protocol.imu.get_orientation(.1, .1, .1, .1)
        acc = self.imu.protocol.imu.get_acc(q)
        self.out_record.append((g, wall_time, q, acc))

        return a.astype(np.uint8)

    def on_end(self) -> Optional['GameState']:
        with open(f"{self.out_name}.pickle", 'wb') as out_pickle:
            pickle.dump(self.out_record, out_pickle)
        self.out_video.release()
        return super().on_end()


class EasyState(AIRecordingState):
    def __init__(self, cam, grabbed_frame, imu, out_name):
        super().__init__(cam, grabbed_frame, imu, out_name, easy_music, easy_bpm / 4)


class NormalState(AIRecordingState):
    def __init__(self, cam, grabbed_frame, imu, out_name):
        super().__init__(cam, grabbed_frame, imu, out_name, normal_music, normal_bpm / 2)


class HardState(AIRecordingState):
    def __init__(self, cam, grabbed_frame, imu, out_name):
        super().__init__(cam, grabbed_frame, imu, out_name, hard_music, hard_bpm)


class MenuState(GameState):
    def __init__(self, cam, imu, out_name):
        super().__init__(cam, imu, out_name, menu_music, loop=True)
        self.last_frame = None
        self.imu_failure = None
        self.menu_np = menu_np
        self.out_name = out_name
        try:
            if isinstance(imu, IMU):
                self.pro = imu
            if imu is not None and imu is not False:
                if imu is True:
                    com = get_coms_in_range()
                    if len(com) == 0:
                        raise RuntimeError("No COMs found.")
                    com = com[0]
                else:
                    com = imu

                con, self.pro = get_i2c_imu(com)
                self.pro.start()
        except Exception as e:
            self.imu_failure = e
            self.menu_np = cv2.putText(self.menu_np, f'IMU Failed to initialize:  {e}',
                                       (0, 24), cv2.FONT_HERSHEY_SIMPLEX, .5, (0, 0, 255, 255), 1)

    def preprocess_video(self, video: np.ndarray) -> np.ndarray:
        self.last_frame = video
        return super().preprocess_video(video)

    def on_video(self, video):
        a = overlay_transparent(video, self.menu_np)
        return a

    def on_mouse(self, mouse_event):
        overlay_button_positions = {
            "easy": [[64, 128], [64 + 128, 128 + 128]],
            "normal": [[64 * 4, 64 * 2], [64 * 4 + 128, 64 * 2 + 128]],
            "hard": [[64 * 7, 64 * 2], [64 * 7 + 128, 64 * 2 + 128]]
        }
        button_commands = {
            "easy": EasyState,
            "normal": NormalState,
            "hard": HardState,
        }

        if self.video_shape is not None and mouse_event is not None and mouse_event.flags == cv2.EVENT_FLAG_LBUTTON:
            h, w = menu_np.shape[0], menu_np.shape[1]
            v_h, v_w = self.video_shape[0], self.video_shape[1]

            top_add = (v_h - h) / 2.0
            left_add = (v_w - w) / 2.0

            for key, value in overlay_button_positions.items():
                if value[0][0] < mouse_event.x - left_add < value[1][0] and \
                        value[0][1] < mouse_event.y - top_add < value[1][1]:
                    self.next_state = button_commands[key](self.cam, np.flip(self.last_frame, 1), self.pro,
                                                           self.out_name)


def run_game(start_state, led_tracker=True, imu: Optional[Union[str, bool]] = True, out_name: str = "train"):
    from virtualreality.calibration.manual_color_mask_calibration import list_supported_capture_properties
    import logging
    cam = 0
    vs = cv2.VideoCapture(cam)
    vs.set(cv2.CAP_PROP_EXPOSURE, -7)
    vs_supported = list_supported_capture_properties(vs)
    if led_tracker:
        if "CAP_PROP_AUTO_EXPOSURE" not in vs_supported:
            logging.warning(f"Camera {cam} does not support turning on/off auto exposure.")
        else:
            vs.set(cv2.CAP_PROP_AUTO_EXPOSURE, 0.25)

        if "CAP_PROP_EXPOSURE" not in vs_supported:
            logging.warning(f"Camera {cam} does not support directly setting exposure.")
        else:
            vs.set(cv2.CAP_PROP_EXPOSURE, -7)
    else:
        if "CAP_PROP_AUTO_EXPOSURE" not in vs_supported:
            logging.warning(f"Camera {cam} does not support turning on/off auto exposure.")
        else:
            vs.set(cv2.CAP_PROP_AUTO_EXPOSURE, 100.0)

        vs.release()

    current_state = start_state(0, imu, out_name)
    while current_state is not None:
        current_state = current_state.run()


if __name__ == "__main__":
    run_game(MenuState)



"""Templates for pose estimators, or posers."""
import asyncio
import numbers
import warnings

from ..util import utilz as u
from .template_base import *
from .poses import *

class PoserTemplate(PoserTemplateBase):
    """
    poser template.

    supplies 3 tracked device pose dicts:
        self.pose - hmd pose, position, orientation, velocity and angular velocity
        self.pose_controller_r - same as hmd pose +controller inputs for controller 1
        self.pose_controller_l - same as hmd pose +controller inputs for controller 2

    for more info: help(PoserTemplateBase)

    Example:
        class MyPoser(PoserTemplate):
            def __init__(self, *args, **kwargs):
                super().__init__(**kwargs)

            @PoserTemplate.register_member_thread(0.5)
            async def example_thread(self):
                while self.coro_keep_alive['example_thread'].is_alive:
                    self.pose.x += 0.04

                    await asyncio.sleep(self.coro_keep_alive['example_thread'].sleep_delay)

        poser = MyPoser()

        asyncio.run(poser.main())

    more examples:
        https://github.com/okawo80085/hobo_vr/blob/master/virtualreality/trackers/color_tracker.py
        https://github.com/okawo80085/hobo_vr/blob/master/examples/poserTemplate.py
        https://github.com/okawo80085/hobo_vr/blob/master/examples/poserClient.py

    """

    def __init__(self, *args, **kwargs):
        """
        init

        :addr: is the address of the server to connect to, stored in self.addr
        :port: is the port of the server to connect to, stored in self.port
        :send_delay: sleep delay for the self.send thread(in seconds)
        :recv_delay: sleep delay for the self.recv thread(in seconds)
        """
        super().__init__(**kwargs)

        self.pose = Pose()
        self.pose_controller_r = ControllerState(pose=(0.5, 1, -1, 1, 0, 0, 0))
        self.pose_controller_l = ControllerState(pose=(0.5, 1.1, -1, 1, 0, 0, 0))

    async def send(self):
        """Send all poses thread."""
        while self.coro_keep_alive["send"].is_alive:
            try:
                msg = u.format_str_for_write(
                    " ".join(
                        [str(i) for i in get_slot_values(self.pose)]
                        + [str(i) for i in get_slot_values(self.pose_controller_r)]
                        + [str(i) for i in get_slot_values(self.pose_controller_l)]
                    )
                )

                self.writer.write(msg)
                await self.writer.drain()

                await asyncio.sleep(self.coro_keep_alive["send"].sleep_delay)
            except Exception as e:
                print(f"send failed: {e}")
                self.coro_keep_alive["send"].is_alive = False
                break


class PoserClient(PoserTemplate, PoserClientBase):
    """
    PoserClient.

    example usage:
        poser = PoserClient()

        @poser.thread_register(1)
        async def lol():
            while poser.coro_keep_alive['lol'].is_alive:
                poser.pose.x += 0.2

                await asyncio.sleep(poser.coro_keep_alive['lol'].sleep_delay)

        asyncio.run(poser.main())
    """

    def __init__(self, *args, **kwargs):
        """init"""
        print ('starting the client')
        super().__init__(*args, **kwargs)



"""
pyvr track.

Usage:
    pyvr track [options]

Options:
   -h, --help
   -c, --camera <camera>                    Source of the camera to use for calibration [default: 0]
   -r, --resolution <res>                   (in progress) Input resolution in width and height [default: -1x-1]
   -l, --load_calibration <file>            (optional) Load color mask calibration settings
   -m, --load_calibration_map <map_file>    (optional) Load color mask calibration map settings
   -i, --ip_address <ip_address>            IP Address of the server to connect to [default: 127.0.0.1]
   -s, --standalone                         Run the server alongside the tracker.
"""

# undocumented reference
# kinda specific to okawo's hardware

import asyncio
import math
import sys
from copy import copy
import numpy as np
import serial
import serial.threaded
import pyrr
from pyrr import Quaternion
from docopt import docopt
import glob

from ..util import utilz as u
from .. import __version__
from .. import templates
from ..calibration.manual_color_mask_calibration import CalibrationData
from ..calibration.manual_color_mask_calibration import colordata_to_blob
from ..calibration.manual_color_mask_calibration import load_mapdata_from_file
from ..server import server
from ..templates import ControllerState

def check_serial_dict(ser_dict, key):
    if key in ser_dict:
        if ser_dict[key] in glob.glob("/dev/tty[A-Za-z]*"):
            return True

    return False

class Poser(templates.PoserTemplate):
    """A pose estimator."""

    def __init__(self, *args, camera=4, width=-1, height=-1, calibration_file=None, calibration_map_file=None, **kwargs):
        """Create a pose estimator."""
        super().__init__(*args, **kwargs)

        self.temp_pose = ControllerState()

        self.serialPaths = {"cont_r": "/dev/ttyUSB2","cont_l": "/dev/ttyUSB1", "hmd": "/dev/ttyUSB0"}

        self.mode = 0
        self._serialResetYaw = False
        self.usePos = True

        self.camera = camera
        self.width = width
        self.height = height

        if calibration_file is not None:
            try:
                if calibration_map_file is not None:
                    self.calibration = colordata_to_blob(CalibrationData.load_from_file(calibration_file), load_mapdata_from_file(calibration_map_file))

                else:
                    raise Exception('no map file provided')

            except Exception as e:
                print (f'calibration file provided, but {repr(e)}, ignoring calibration file..')
                self.calibration = {
                                "blue": {"h": (98, 10), "s": (200, 55), "v": (250, 32)},
                                "green": {"h": (68, 15), "s": (135, 53), "v": (255, 50)},
                                    }

        else:
            self.calibration = {
                            "blue": {"h": (98, 10), "s": (200, 55), "v": (250, 32)},
                            "green": {"h": (68, 15), "s": (135, 53), "v": (255, 50)},
                                }

        print (f'current color masks {self.calibration}')


    @templates.PoserTemplate.register_member_thread(1 / 90)
    async def mode_switcher(self):
        """Check to switch between left and right controllers."""
        while self.coro_keep_alive["mode_switcher"].is_alive:
            try:
                if self.mode == 1:
                    self.pose_controller_r.trackpad_touch = self.temp_pose.trackpad_touch
                    self.pose_controller_r.trackpad_click = self.temp_pose.trackpad_click
                    self.pose_controller_r.trackpad_x = self.temp_pose.trackpad_x
                    self.pose_controller_r.trackpad_y = self.temp_pose.trackpad_y
                    self.pose_controller_r.trigger_value = self.temp_pose.trigger_value
                    self.pose_controller_r.trigger_click = self.temp_pose.trigger_click
                    self.pose_controller_r.system = self.temp_pose.system
                    self.pose_controller_r.grip = self.temp_pose.grip
                    self.pose_controller_r.menu = self.temp_pose.menu

                else:
                    self.pose_controller_l.trackpad_touch = self.temp_pose.trackpad_touch
                    self.pose_controller_l.trackpad_click = self.temp_pose.trackpad_click
                    self.pose_controller_l.trackpad_x = self.temp_pose.trackpad_x
                    self.pose_controller_l.trackpad_y = self.temp_pose.trackpad_y
                    self.pose_controller_l.trigger_value = self.temp_pose.trigger_value
                    self.pose_controller_l.trigger_click = self.temp_pose.trigger_click
                    self.pose_controller_l.system = self.temp_pose.system
                    self.pose_controller_l.grip = self.temp_pose.grip
                    self.pose_controller_l.menu = self.temp_pose.menu

                await asyncio.sleep(self.coro_keep_alive["mode_switcher"].sleep_delay)

            except Exception as e:
                print(f"failed mode_switcher: {e}")
                break
        self.coro_keep_alive["mode_switcher"].is_alive = False

    @templates.PoserTemplate.register_member_thread(1 / 65)
    async def get_location(self):
        """Get locations from blob trackers."""
        try:
            offsets = [0.6981317007977318, 0, 0]
            t1 = u.BlobTracker(
                self.camera,
                color_masks=self.calibration,
            )

        except Exception as e:
            print(f"failed to init get_location: {e}")
            self.coro_keep_alive["get_location"].is_alive = False
            return

        with t1:
            while self.coro_keep_alive["get_location"].is_alive:
                try:
                    poses = t1.get_poses()

                    u.rotate(poses, offsets)

                    if self.usePos:
                        self.pose_controller_l.x = round(-poses["green"]["x"] * 0.9, 6)
                        self.pose_controller_l.y = round(poses["green"]["y"] * 0.85, 6)
                        self.pose_controller_l.z = round(poses["green"]["z"] * 1.8, 6)

                        self.pose_controller_r.x = round(-poses["yellow"]["x"] * 0.9, 6)
                        self.pose_controller_r.y = round(poses["yellow"]["y"] * 0.85, 6)
                        self.pose_controller_r.z = round(poses["yellow"]["z"] * 1.8, 6)

                        self.pose.x = round(-poses["blue"]["x"] * 0.9 - 0.01, 6)
                        self.pose.y = round(poses["blue"]["y"] * 0.85 - 0.07, 6)
                        self.pose.z = round(poses["blue"]["z"] * 2 + 0.05, 6)

                    await asyncio.sleep(self.coro_keep_alive["get_location"].sleep_delay)

                except Exception as e:
                    print("stopping get_location:", e)
                    break

        self.coro_keep_alive["get_location"].is_alive = False

    @templates.PoserTemplate.register_member_thread(1 / 100)
    async def serial_listener2(self):
        """Get controller data from serial."""
        irl_rot_off = Quaternion.from_z_rotation(np.pi/2) # imu on this controller is rotated 90 degrees irl for me

        my_off = Quaternion()

        if not check_serial_dict(self.serialPaths, 'cont_l'):
            print ('failed to init serial_listener2: bad serial dict for key \'cont_l\'')
            self.coro_keep_alive["serial_listener2"].is_alive = False
            return

        with serial.Serial(self.serialPaths["cont_l"], 115200, timeout=1 / 4) as ser:
            with serial.threaded.ReaderThread(ser, u.SerialReaderFactory) as protocol:
                protocol.write_line("nut")
                await asyncio.sleep(5)

                while self.coro_keep_alive["serial_listener2"].is_alive:
                    try:
                        gg = u.get_numbers_from_text(protocol.last_read, ',')

                        if len(gg) > 0:
                            (w, x, y, z, trgr, grp, util, sys, menu, padClk, padY, padX,) = gg

                            my_q = Quaternion([-y, z, -x, w])

                            my_q = Quaternion(my_off * my_q * irl_rot_off).normalised
                            self.pose_controller_l.r_w = round(my_q[3], 5)
                            self.pose_controller_l.r_x = round(my_q[0], 5)
                            self.pose_controller_l.r_y = round(my_q[1], 5)
                            self.pose_controller_l.r_z = round(my_q[2], 5)

                            self.temp_pose.trigger_value = trgr
                            self.temp_pose.grip = grp
                            self.temp_pose.menu = menu
                            self.temp_pose.system = sys
                            self.temp_pose.trackpad_click = padClk

                            self.temp_pose.trackpad_x = round((padX - 428) / 460, 3) * (-1)
                            self.temp_pose.trackpad_y = round((padY - 530) / 540, 3) * (-1)

                            self._serialResetYaw = False
                            if self.temp_pose.trackpad_x > 0.6 and util:
                                self.mode = 1

                            elif self.temp_pose.trackpad_x < -0.6 and util:
                                self.mode = 0

                            elif self.temp_pose.trackpad_y > 0.6 and util:
                                self.usePos = True

                            elif self.temp_pose.trackpad_y < -0.6 and util:
                                self.usePos = False

                            elif util:
                                my_off = Quaternion([0, z, 0, w]).inverse.normalised
                                self._serialResetYaw = True

                        if abs(self.temp_pose.trackpad_x) > 0.09 or abs(self.temp_pose.trackpad_y) > 0.09:
                            self.temp_pose.trackpad_touch = 1

                        else:
                            self.temp_pose.trackpad_touch = 0

                        if self.temp_pose.trigger_value > 0.9:
                            self.temp_pose.trigger_click = 1

                        else:
                            self.temp_pose.trigger_click = 0

                        await asyncio.sleep(self.coro_keep_alive["serial_listener2"].sleep_delay)

                    except Exception as e:
                        print(f"{self.serial_listener2.__name__}: {e}")
                        break
        self.coro_keep_alive["serial_listener2"].is_alive = False

    @templates.PoserTemplate.register_member_thread(1 / 100)
    async def serial_listener3(self):
        """Get orientation data from serial."""
        irl_rot_off = Quaternion.from_x_rotation(np.pi/3) # imu on this controller is rotated 90 degrees irl for me

        my_off = Quaternion()

        if not check_serial_dict(self.serialPaths, 'cont_r'):
            print ('failed to init serial_listener3: bad serial dict for key \'cont_r\'')
            self.coro_keep_alive["serial_listener3"].is_alive = False
            return

        with serial.Serial(self.serialPaths["cont_r"], 115200, timeout=1 / 5) as ser2:
            with serial.threaded.ReaderThread(ser2, u.SerialReaderFactory) as protocol:
                protocol.write_line("nut")
                await asyncio.sleep(5)

                while self.coro_keep_alive["serial_listener3"].is_alive:
                    try:
                        gg = u.get_numbers_from_text(protocol.last_read, ',')

                        if len(gg) > 0:
                            w, x, y, z = gg
                            my_q = Quaternion([-y, z, -x, w])

                            if self._serialResetYaw:
                                my_off = Quaternion([0, z, 0, w]).inverse.normalised

                            my_q = Quaternion(my_off * my_q * irl_rot_off).normalised
                            self.pose_controller_r.r_w = round(my_q[3], 5)
                            self.pose_controller_r.r_x = round(my_q[0], 5)
                            self.pose_controller_r.r_y = round(my_q[1], 5)
                            self.pose_controller_r.r_z = round(my_q[2], 5)

                        await asyncio.sleep(self.coro_keep_alive["serial_listener3"].sleep_delay)

                    except Exception as e:
                        print(f"{self.serial_listener.__name__}: {e}")
                        break
        self.coro_keep_alive["serial_listener3"].is_alive = False

    @templates.PoserTemplate.register_member_thread(1 / 100)
    async def serial_listener(self):
        """Get orientation data from serial."""
        my_off = Quaternion()

        if not check_serial_dict(self.serialPaths, 'hmd'):
            print ('failed to init serial_listener: bad serial dict for key \'hmd\'')
            self.coro_keep_alive["serial_listener"].is_alive = False
            return


        with serial.Serial(self.serialPaths["hmd"], 115200, timeout=1 / 5) as ser2:
            with serial.threaded.ReaderThread(ser2, u.SerialReaderFactory) as protocol:
                protocol.write_line("nut")
                await asyncio.sleep(5)

                while self.coro_keep_alive["serial_listener"].is_alive:
                    try:
                        gg = u.get_numbers_from_text(protocol.last_read, ',')

                        if len(gg) > 0:
                            w, x, y, z = gg
                            my_q = Quaternion([-y, z, -x, w])

                            if self._serialResetYaw:
                                my_off = Quaternion([0, z, 0, w]).inverse.normalised

                            my_q = Quaternion(my_off * my_q).normalised
                            self.pose.r_w = round(my_q[3], 5)
                            self.pose.r_x = round(my_q[0], 5)
                            self.pose.r_y = round(my_q[1], 5)
                            self.pose.r_z = round(my_q[2], 5)

                        await asyncio.sleep(self.coro_keep_alive["serial_listener"].sleep_delay)

                    except Exception as e:
                        print(f"{self.serial_listener.__name__}: {e}")
                        break
        self.coro_keep_alive["serial_listener"].is_alive = False

def run_poser_only(addr="127.0.0.1", cam=4, colordata=None, mapdata=None):
    """Run the poser only. The server must be started in another program."""
    t = Poser(addr=addr, camera=cam, calibration_file=colordata, calibration_map_file=mapdata)
    asyncio.run(t.main())


def run_poser_and_server(addr="127.0.0.1", cam=4, colordata=None, mapdata=None):
    """Run the poser and server in one program."""
    t = Poser(addr=addr, camera=cam, calibration_file=colordata, calibration_map_file=mapdata)
    server.run_til_dead(t)


def main():
    """Run color tracker entry point."""
    # allow calling from both python -m and from pyvr:
    argv = sys.argv[1:]
    if sys.argv[1] != "track":
        argv = ["track"] + argv

    args = docopt(__doc__, version=f"pyvr version {__version__}", argv=argv)

    width, height = args["--resolution"].split("x")

    print (args)

    if args["--camera"].isdigit():
        cam = int(args["--camera"])
    else:
        cam = args["--camera"]

    if args["--standalone"]:
        run_poser_and_server(args["--ip_address"], cam, args['--load_calibration'], args['--load_calibration_map'])
    else:
        run_poser_only(args["--ip_address"], cam, args['--load_calibration'], args['--load_calibration_map'])



"""Virtualreality Utility Module"""
from ..logging import log

logger = log.setup_custom_logger(name=__name__, 
    level="INFO",
    file="../logs/util.log")

logger.debug("created util.log log file")


"""Calibration functions."""
from ..logging import log

logger = log.setup_custom_logger(name=__name__,
    level="INFO",
    file="../logs/calibration.log",
    console_logging=False)

logger.debug("created calibration.log log file")



"""
pyvr track.

Usage:
    pyvr track [options]

Options:
   -h, --help
   -l, --load_calibration <file>    (in progress) Load color mask calibration settings [default: ranges.pickle]
   -i, --ip_address <ip_address>    IP Address of the server to connect to [default: 127.0.0.1]
   --calibration-data <calib_file>  Source of the camera to use for calibration
   -s, --standalone                 Run the server alongside the tracker.
"""

import asyncio
import math
import sys

import serial
import serial.threaded
import pyrr
from docopt import docopt
from ..util import utilz as u
from ..util.IMU import get_i2c_imu
from ..util.kalman import EulerKalman
import numpy as np
from .. import __version__
from .. import templates
from ..server import server
from ..templates import PoseEuler
from ..calibration.manual_color_mask_calibration import CalibrationData, ColorRange


def horizontal_asymptote(val, max_val, speed=1):
    b = max_val + 1
    out_val = (speed * b * val + b) / (speed * val + b) - 1
    return out_val


class HeadsetPoserPiece(object):
    def __init__(self, tracker_camera=0, imu_factory=get_i2c_imu, blob_colors={}, translation_offsets={}):
        self.tracker_camera = tracker_camera
        self.conn, self.imu = imu_factory()
        self.pose = PoseEuler()
        self.blob_colors = blob_colors
        self.translation_offsets = translation_offsets
        self._last_blob_update = 0
        self._last_imu_update = 0
        self.kalman = EulerKalman()

        assert len(blob_colors.keys()) == len(translation_offsets.keys())

        try:
            if self.blob_colors:
                self.blob_tracker = u.BlobTracker(self.tracker_camera, offsets=[0, 0, 0], color_masks=self.blob_colors,)
            else:
                self.blob_tracker = None
        except Exception as e:
            print(f"failed to init headset blob tracker: {repr(e)}")
            self.blob_tracker = None

    def update_blob_xyz(self):
        try:
            if self.blob_colors and self.blob_tracker.time_of_last_frame != self._last_blob_update:
                self._last_blob_update = self.blob_tracker.time_of_last_frame
                pose_list = []
                poses = self.blob_tracker.get_poses()
                for color in self.blob_colors.keys():
                    pose_x = poses[color]["x"]
                    pose_y = poses[color]["y"]
                    pose_z = poses[color]["z"]
                    pose_x += self.translation_offsets[color]["x"]
                    pose_y += self.translation_offsets[color]["y"]
                    pose_z += self.translation_offsets[color]["z"]
                    pose_list.append((pose_x, pose_y, pose_z))
                pose_x = sum((p[0] for p in pose_list)) / len(pose_list)
                pose_y = sum((p[1] for p in pose_list)) / len(pose_list)
                pose_z = sum((p[2] for p in pose_list)) / len(pose_list)
                self.kalman.apply_blob((pose_x, pose_y, pose_z))
        except Exception as e:
            print("Could not get blob location:", e)

    def update_imu_info(self):
        if self.imu.protocol is not None and self.imu.protocol.time_of_last_data != self._last_imu_update:
            self._last_imu_update = self.imu.protocol.time_of_last_data
            self.kalman.apply_imu(self.imu.protocol)


class Poser(templates.PoserTemplate):
    """A pose estimator."""

    def __init__(self, *args, camera=4, width=-1, height=-1, calibration_file=None, **kwargs):
        """Create a pose estimator."""
        super().__init__(*args, **kwargs)

        self.headset = HeadsetPoserPiece(
            0,
            blob_colors={
                "blue": {"h": (98, 10), "s": (200, 55), "v": (250, 32)},
                "green": {"h": (68, 15), "s": (135, 53), "v": (255, 50)},
            },
            translation_offsets={"blue": [0, 0, 0], "green": [0, 0, 0]},
        )

        self.mode = 0
        self.useVelocity = False

        self.camera = camera
        self.width = width
        self.height = height

    @templates.PoserTemplate.register_member_thread(1 / 60)
    async def get_location(self):
        """Get locations from blob trackers."""
        try:
            t1 = u.BlobTracker(
                self.camera,
                offsets=[0.6981317007977318, 0, 0],
                color_masks={
                    "blue": {"h": (98, 10), "s": (200, 55), "v": (250, 32)},
                    "green": {"h": (68, 15), "s": (135, 53), "v": (255, 50)},
                },
            )

        except Exception as e:
            print(f"failed to init location tracker: {repr(e)}")
            return

        with t1:
            while self.coro_keep_alive["get_location"].is_alive:
                try:
                    poses = t1.get_poses()

                    self.temp_pose.x = round(-poses["green"]["x"], 6)
                    self.temp_pose.y = round(poses["green"]["y"] + 1, 6)
                    self.temp_pose.z = round(poses["green"]["z"], 6)

                    self.pose.x = round(-poses["blue"]["x"] - 0.01, 6)
                    self.pose.y = round(poses["blue"]["y"] + 1 - 0.07, 6)
                    self.pose.z = round(poses["blue"]["z"] - 0.03, 6)

                    await asyncio.sleep(self.coro_keep_alive["get_location"].sleep_delay)

                except Exception as e:
                    print("stopping get_location:", e)
                    break

    @templates.PoserTemplate.register_member_thread(1 / 100)
    async def serial_listener_2(self):
        """Get controller data from serial."""
        past_velocity = [0, 0, 0]
        velocity_until_reset = 0
        temp_for_offz = {"x": 0, "y": 0, "z": 0}
        while self.coro_keep_alive["serial_listener_2"].is_alive:
            try:
                yaw_offset = 0
                with serial.Serial(self.serialPaths["green"], 115200, timeout=1 / 4) as ser:
                    with serial.threaded.ReaderThread(ser, u.SerialReaderFactory) as protocol:
                        for _ in range(10):
                            protocol.write_line("nut")
                            await asyncio.sleep(1)

                        while self.coro_keep_alive["serial_listener_2"].is_alive:
                            try:
                                gg = u.get_numbers_from_text(protocol.last_read)

                                if len(gg) > 0:
                                    (y, p, r, az, ax, ay, trgr, grp, util, sys, menu, padClk, padY, padX,) = gg

                                    if velocity_until_reset < 20:

                                        u.rotate_y(
                                            {"": temp_for_offz}, math.radians(-yaw_offset),
                                        )
                                        past_velocity = [
                                            temp_for_offz["x"],
                                            temp_for_offz["y"],
                                            temp_for_offz["z"],
                                        ]
                                        temp_for_offz["x"] = round(
                                            past_velocity[0]
                                            - ax * self.coro_keep_alive["serial_listener_2"].sleep_delay * 0.005,
                                            4,
                                        )
                                        temp_for_offz["y"] = round(
                                            past_velocity[1]
                                            - ay * self.coro_keep_alive["serial_listener_2"].sleep_delay * 0.005,
                                            4,
                                        )
                                        temp_for_offz["z"] = round(
                                            past_velocity[2]
                                            - az * self.coro_keep_alive["serial_listener_2"].sleep_delay * 0.005,
                                            4,
                                        )
                                        u.rotate_y(
                                            {"": temp_for_offz}, math.radians(yaw_offset),
                                        )

                                        if self.useVelocity:
                                            self.temp_pose.vel_x = temp_for_offz["x"]
                                            self.temp_pose.vel_y = temp_for_offz["y"]
                                            self.temp_pose.vel_z = temp_for_offz["z"]

                                        velocity_until_reset += 1

                                    else:
                                        velocity_until_reset = 0
                                        temp_for_offz = {"x": 0, "y": 0, "z": 0}
                                        self.temp_pose.vel_x = 0
                                        self.temp_pose.vel_y = 0
                                        self.temp_pose.vel_z = 0

                                    yaw = y
                                    eulers = [y + yaw_offset, p, r]
                                    eulers = np.radians(eulers)
                                    x, y, z, w = pyrr.Quaternion.from_eulers(eulers).xyzw

                                    self.temp_pose.r_w = round(w, 4)
                                    self.temp_pose.r_x = round(y, 4)
                                    self.temp_pose.r_y = round(-x, 4)
                                    self.temp_pose.r_z = round(-z, 4)

                                    self.temp_pose.trigger_value = abs(trgr - 1)
                                    self.temp_pose.grip = abs(grp - 1)
                                    self.temp_pose.menu = abs(menu - 1)
                                    self.temp_pose.system = abs(sys - 1)
                                    self.temp_pose.trackpad_click = abs(padClk - 1)

                                    self.temp_pose.trackpad_x = round((padX - 524) / 530, 3) * (-1)
                                    self.temp_pose.trackpad_y = round((padY - 506) / 512, 3) * (-1)

                                    temp_mode = abs(util - 1)

                                    self._serialResetYaw = False
                                    if self.temp_pose.trackpad_x > 0.6 and temp_mode:
                                        self.mode = 1

                                    elif self.temp_pose.trackpad_x < -0.6 and temp_mode:
                                        self.mode = 0

                                    elif self.temp_pose.trackpad_y < -0.6 and temp_mode:
                                        self.useVelocity = False

                                    elif self.temp_pose.trackpad_y > 0.6 and temp_mode:
                                        self.useVelocity = True

                                    elif temp_mode:
                                        yaw_offset = 0 - yaw
                                        self._serialResetYaw = True

                                if abs(self.temp_pose.trackpad_x) > 0.07 or abs(self.temp_pose.trackpad_y) > 0.07:
                                    self.temp_pose.trackpad_touch = 1

                                else:
                                    self.temp_pose.trackpad_touch = 0

                                if self.temp_pose.trigger_value > 0.9:
                                    self.temp_pose.trackpad_click = 1

                                else:
                                    self.temp_pose.trigger_click = 0

                                await asyncio.sleep(self.coro_keep_alive["serial_listener_2"].sleep_delay)

                            except Exception as e:
                                print(f"{self.serial_listener_2.__name__}: {e}")
                                break

            except Exception as e:
                print(f"serial_listener_2: {e}")

            await asyncio.sleep(1)

    @templates.PoserTemplate.register_member_thread(1 / 100)
    async def serial_listener(self):
        """Get orientation data from serial."""
        while self.coro_keep_alive["serial_listener"].is_alive:
            try:
                with serial.Serial(self.serialPaths["blue"], 115200, timeout=1 / 5) as ser2:
                    with serial.threaded.ReaderThread(ser2, u.SerialReaderFactory) as protocol:
                        yaw_offset = 0
                        for _ in range(10):
                            protocol.write_line("nut")
                            await asyncio.sleep(1)

                        while self.coro_keep_alive["serial_listener"].is_alive:
                            try:
                                gg = u.get_numbers_from_text(protocol.last_read)

                                if len(gg) > 0:
                                    ypr = gg

                                    eulers = [ypr[0] + yaw_offset, ypr[1], ypr[2]]
                                    eulers = np.radians(eulers)
                                    x, y, z, w = pyrr.Quaternion.from_eulers(eulers).xyzw

                                    # self.pose['rW'] = rrr2[0]
                                    # self.pose['rX'] = -rrr2[2]
                                    # self.pose['rY'] = rrr2[3]
                                    # self.pose['rZ'] = -rrr2[1]

                                    self.pose.r_w = round(w, 4)
                                    self.pose.r_x = round(y, 4)
                                    self.pose.r_y = round(-x, 4)
                                    self.pose.r_z = round(-z, 4)

                                    if self._serialResetYaw:
                                        yaw_offset = 0 - ypr[0]

                                await asyncio.sleep(self.coro_keep_alive["serial_listener"].sleep_delay)

                            except Exception as e:
                                print(f"{self.serial_listener.__name__}: {e}")
                                break

            except Exception as e:
                print(f"serial_listener: {e}")

            await asyncio.sleep(1)


def run_poser_only(addr, calib):
    """Run the poser only. The server must be started in another program."""
    t = Poser(addr=addr, calib=calib)
    asyncio.run(t.main())


def run_poser_and_server(addr, calib):
    """Run the poser and server in one program."""
    t = Poser(addr=addr, calib=calib)
    server.run_til_dead(t)


def main():
    """Run color tracker entry point."""
    # allow calling from both python -m and from pyvr:
    argv = sys.argv[1:]
    if len(argv) < 2 or sys.argv[1] != "track":
        argv = ["track"] + argv

    args = docopt(__doc__, version=f"pyvr version {__version__}", argv=argv)

    if args["--calibration-data"]:
        calib = CalibrationData.load_from_file(args["--calibration-data"])
    else:
        calib = CalibrationData.load_from_file('rgby_calibration.pickle')

    if args["--standalone"]:
        run_poser_and_server(args["--ip_address"], calib)
    else:
        run_poser_only(args["--ip_address"], calib)

    print(args)



"""Virtual reality tools and drivers for python."""
from . import util
from . import templates

from .logging import log

logger = log.setup_custom_logger(name=__name__,
    level="INFO",
    file="./logs/app.log")

logger.debug("created app.log log file")

__version__ = "0.0.3"


"""Templates for pose estimators, or posers. unlimited devices upgrade version."""
import asyncio
import numbers
import warnings

from ..util import utilz as u
from .template_base import *
from .poses import *
import re


class UduPoserTemplate(PoserTemplateBase):
    """
    udu poser template.

    supplies a list of poses:
        self.poses - pose object will correspond to value of device_list_manifest, keep in mind that, as of now, only 3 types of devices are supported(h-hmd, c-controller, t-tracker)

    for more info: help(PoserTemplateBase)

    Example:
        class MyPoser(UduPoserTemplate):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            @UduPoserTemplate.register_member_thread(0.5)
            async def example_thread(self):
                while self.coro_keep_alive['example_thread'].is_alive:
                    self.poses[0].x += 0.04

                    await asyncio.sleep(self.coro_keep_alive['example_thread'].sleep_delay)

        poser = MyPoser('h c c')

        asyncio.run(poser.main())

    more examples:
        https://github.com/okawo80085/hobo_vr/blob/master/examples/uduPoserClient.py

    """

    def __init__(self, device_list_manifest, *args, **kwargs):
        """
        init

        :device_list_manifest: should match this regex: ([htc][ ])+([htc]$)|[htc]$
        :addr: is the address of the server to connect to, stored in self.addr
        :port: is the port of the server to connect to, stored in self.port
        :send_delay: sleep delay for the self.send thread(in seconds)
        :recv_delay: sleep delay for the self.recv thread(in seconds)
        """
        super().__init__(**kwargs)

        re_s = re.search('([htc][ ])+([htc]$)|[htc]$', device_list_manifest)

        if not device_list_manifest or re_s is None:
            raise RuntimeError('empty pose struct')

        if re_s.group() != device_list_manifest:
            raise RuntimeError(f'invalid pose struct: {repr(device_list_manifest)}')

        self.poses = []
        self.device_types = device_list_manifest.split(' ')

        new_struct = []
        for i in self.device_types:
            if i == 'h' or i == 't':
                self.poses.append(Pose())
                new_struct.append(f'{i}{len(self.poses[-1])}')

            elif i == 'c':
                self.poses.append(ControllerState())
                new_struct.append(f'{i}{len(self.poses[-1])}')

        new_struct = ' '.join(new_struct)

        print (f'total of {len(self.poses)} device(s) have been added, a new pose struct has been generated: {repr(new_struct)}')
        print ('full device list is now available through self.poses')

    async def send(self):
        """Send all poses thread."""
        poses_index = range(len(self.device_types))
        while self.coro_keep_alive["send"].is_alive:
            try:
                msg = u.format_str_for_write(' '.join([str(j) for i in poses_index for j in get_slot_values(self.poses[i])]))

                self.writer.write(msg)
                await self.writer.drain()

                await asyncio.sleep(self.coro_keep_alive["send"].sleep_delay)
            except Exception as e:
                print(f"send failed: {e}")
                self.coro_keep_alive["send"].is_alive = False
                break


class UduPoserClient(UduPoserTemplate, PoserClientBase):
    """
    UduPoserClient.

    example usage:
        poser = UduPoserClient('h c c')

        @poser.thread_register(1)
        async def lol():
            while poser.coro_keep_alive['lol'].is_alive:
                poser.poses[0].x += 0.2

                await asyncio.sleep(poser.coro_keep_alive['lol'].sleep_delay)

        asyncio.run(poser.main())
    """

    def __init__(self, *args, **kwargs):
        """init"""
        print ('starting udu client')
        super().__init__(*args, **kwargs)



from setuptools import setup, find_packages
from virtualreality import __version__

with open("pypi_readme.md", "r") as f:
    long_description = f.read()

setup(
    name="virtualreality",
    version=__version__,
    author="Okawo <okawo.198@gmail.com>, SimLeek <simulator.leek@gmail.com>",
    author_email="",
    description="python side of hobo_vr",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/okawo80085/hobo_vr",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3.7",
        "Topic :: Games/Entertainment",
        "Topic :: Multimedia :: Graphics :: Viewers",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: System :: Hardware :: Hardware Drivers",
    ],
    install_requires=["keyboard", "opencv-python", "numpy", "pykalman", "pyserial", "docopt", "pyrr", "scipy", "displayarray"],
    extras_require={"image": ["imutils"]},
    entry_points={"console_scripts": ["pyvr=virtualreality.main:main"]},
    python_requires='>=3.7',
)



"""Virtualreality Pose Server"""
from . import server
from ..logging import log

__version__ = "0.2" # the server is at this version

logger = log.setup_custom_logger(name=__name__, 
    level="INFO",
    file="../logs/server.log",
    console_logging=True)

logger.debug("created server.log log file")


from unittest import mock


class MockReadUpdates(object):
    def __init__(self, cam, frames, prepend=""):
        self.cam = cam
        self.frames = frames
        self.patches = []
        self.mocks = []
        self.prepend = prepend

    def __enter__(self):
        self.patches.append(mock.patch(self.prepend + "displayarray.window.subscriber_windows.display"))

        for m in self.patches:
            self.mocks.append(m.__enter__())

        main = self.mocks[0].return_value
        main.frames = dict()
        main.frames[str(self.cam)] = self.frames

    def __exit__(self, *args):
        for m in self.patches:
            m.__exit__(*args)

        self.patches = []



"""
pyvr server.

usage:
  server [--show-messages]

options:
    -h --help               shows this message
    -d --show-messages      show messages

"""
from virtualreality.server import server

from . import __version__

from docopt import docopt

args = docopt(__doc__, version=__version__)

my_server = server.Server()
my_server.debug = args["--show-messages"]

server.run_til_dead(conn_handle=my_server)



"""
calculate camera distortion matrix
saves matrix to camera_matrix.pickle

press s to save image from camera, and q to close camera and start calibration

usage:
    pyvr calibrate-cam [options]

Options:
   -c, --camera <camera>                    Source of the camera to use for calibration [default: 0]

"""
import logging
import os
from docopt import docopt
import sys
import pickle

import cv2
import numpy as np

from virtualreality import __version__

images_location = os.path.join(os.path.dirname(__file__), 'calibration_images')
print(images_location)

CHECKERBOARD = (6,9)

def save_images_to_process(cam=0) :
    """save images from specified camera to calibration_images directory for processing

    Keyword Arguments:
        cam {int} -- reference to usb camera input (default: {0})
    """
    img_counter = 1
    cap = cv2.VideoCapture(cam)
    while cap.isOpened() :
        ret, frame = cap.read()
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        
        if not ret :
            logging.error("failed to receive image from camera")
            break

        valid_img, corners = cv2.findChessboardCorners(gray, CHECKERBOARD, cv2.CALIB_CB_ADAPTIVE_THRESH + cv2.CALIB_CB_FAST_CHECK + cv2.CALIB_CB_NORMALIZE_IMAGE)

        if valid_img :
           corners2 = cv2.cornerSubPix(gray, corners, (11,11), (-1,-1), (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001))
           cv2.drawChessboardCorners(frame, CHECKERBOARD, corners2, True)
        
        cv2.imshow('drawn_corners', frame)

        key = cv2.waitKey(1)
        if key & 0xFF == ord('q') or img_counter >= 20 :
            # press q to stop
            cap.release()
            cv2.destroyAllWindows()
            if img_counter == 1 :
                exit
            break
        elif key & 0xFF == ord('s') :
            # press s to save image

            if valid_img:
                print(f"saving image {img_counter}")
                cv2.imwrite(os.path.join(images_location, f"image{img_counter}.jpg"), gray)
                img_counter += 1
            else:
                print("no checkerboard found. Not Saving.")
    
def calibrate_camera(cam=0) :
    """calculate camera matrix from precollected images

    Keyword Arguments:
        cam {int} -- reference to usb camera input (default: {0})
    """
    # Checkerboard dimensions that the method looks for
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)

    object_points = []
    image_points = []

    obj_p = np.zeros((1,CHECKERBOARD[0]*CHECKERBOARD[1], 3), np.float32)
    obj_p[0,:,:2] = np.mgrid[0:CHECKERBOARD[0], 0:CHECKERBOARD[1]].T.reshape(-1,2)

    if not os.path.exists(images_location) :
        os.makedirs(images_location)
        save_images_to_process(0)
    elif len(os.listdir(images_location)) == 0 :
        save_images_to_process(0)

    for image_path in os.listdir(images_location) :

        input_path = os.path.join(images_location, image_path)
        frame = cv2.imread(input_path)
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

        valid_img, corners = cv2.findChessboardCorners(gray, CHECKERBOARD, cv2.CALIB_CB_ADAPTIVE_THRESH + cv2.CALIB_CB_FAST_CHECK + cv2.CALIB_CB_NORMALIZE_IMAGE)

        if valid_img :
           object_points.append(obj_p) 
           corners2 = cv2.cornerSubPix(gray, corners, (11,13), (-1,-1), (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001))
           image_points.append(corners2)
        else :
            logging.warning(f"could not find pattern in {image_path}. Ignoring")

    #calibrate camera with aggregated object_points and image_points from samples
    ret, mtx, dist, rvecs, tvecs = cv2.calibrateCamera(object_points, image_points, (640,480), None, None)

    logging.info("camera matrix: ")
    logging.info(mtx)

    camera_calibration_file = open(os.path.join(os.path.dirname(__file__), "camera_matrix.pickle"), 'wb')
    pickle.dump(mtx, camera_calibration_file)
    camera_calibration_file.close()

def main() :
    # allow calling from both python -m and from pyvr:
    argv = sys.argv[1:]
    if sys.argv[1] != "calibrate-cam":
        argv = ["calibrate-cam"] + argv
    
    args = docopt(__doc__, version=f"pyvr version {__version__}", argv=argv)

    cam = int(args["--camera"]) if args["--camera"].isdigit() else args["--camera"]
    
    calibrate_camera(0)

        

if __name__ == "__main__" :
    main()